import ssl
import atexit
import logging
from typing import Dict, Optional

import grpc

from qm.api.models.debug_data import DebugData
from qm.api.models.grpc_interceptors import DebugInterceptor, AddHeadersInterceptor

logger = logging.getLogger(__name__)


def _create_debug_data_event(debug_data: DebugData, channel: grpc.Channel) -> grpc.Channel:
    """Create debug interceptor for proto-generated gRPC client"""
    # Apply the interceptor to the channel
    intercepted_channel = grpc.intercept_channel(channel, DebugInterceptor(debug_data))
    return intercepted_channel


def _create_add_headers_event(headers: Dict[str, str], channel: grpc.Channel) -> grpc.Channel:
    """Create interceptor to add headers to all gRPC calls"""
    interceptor = AddHeadersInterceptor(headers)
    return grpc.intercept_channel(channel, interceptor)


def create_channel(
    host: str,
    port: int,
    ssl_context: Optional[ssl.SSLContext],
    max_message_size: int,
    headers: Dict[str, str],
    debug_data: Optional[DebugData] = None,
) -> grpc.Channel:
    """
    Create a gRPC channel equivalent to a grpc Channel configuration.

    Args:
        host: Server host
        port: Server port
        max_message_size: Max message size in bytes (used for flow control + message limits)
        ssl_context: Optional ssl.SSLContext. If provided, a secure channel is created.

    Returns:
        grpc.Channel
    """

    address = f"{host}:{port}"

    options = [
        ("grpc.http2.initial_connection_window_size", max_message_size),
        ("grpc.http2.initial_stream_window_size", max_message_size),
        ("grpc.max_receive_message_length", max_message_size),
        ("grpc.max_send_message_length", max_message_size),
    ]

    # ---- TLS channel ----
    if ssl_context is not None:
        # grpc does NOT accept SSLContext directly.
        # We extract what grpc needs from it.

        # Root CAs
        root_certs = None
        if ssl_context.verify_mode != ssl.CERT_NONE:
            try:
                root_certs_list = ssl_context.get_ca_certs(binary_form=True)
                # get_ca_certs returns a list; grpc expects bytes
                root_certs = b"".join(root_certs_list) if root_certs_list else None
            except Exception:
                root_certs = None

        credentials = grpc.ssl_channel_credentials(
            root_certificates=root_certs,
            private_key=None,
            certificate_chain=None,
        )

        channel = grpc.secure_channel(
            address,
            credentials,
            options=options,
        )
    else:
        # ---- Insecure channel ----
        channel = grpc.insecure_channel(
            address,
            options=options,
        )

    if debug_data:
        channel = _create_debug_data_event(debug_data, channel)

    channel = _create_add_headers_event(headers, channel)

    atexit.register(channel.close)

    return channel



from typing import Optional

from google.protobuf.wrappers_pb2 import UInt32Value

from qm._loc import _get_loc
from qm.grpc.qm.pb import inc_qua_pb2
from qm.api.models.capabilities import QopCaps
from qm.qua._expressions import Scalar, to_scalar_pb_expression
from qm.qua._scope_management.scopes_manager import scopes_manager


def ramp_to_zero(element: str, duration: Optional[int] = None) -> None:
    r"""Starting from the last DC value, gradually lowers the DC to zero for `duration` *4nsec

    If `duration` is None, the duration is taken from the element's config

    Warning:
        This feature does not protect from voltage jumps. Those can still occur, i.e. when the data sent to the
        analog output is outside the range -0.5 to $0.5 - 2^{16}$ and thus will have an overflow.

    Args:
        element (str): element for ramp to zero
        duration (Union[int,None]): time , `in multiples of 4nsec`.
            Range: [4, $2^{24}$] in steps of 1, or `None` to take
            value from config
    """
    duration = duration if duration is None else int(duration)
    loc = _get_loc()
    duration_val = UInt32Value(value=duration) if duration else None
    statement = inc_qua_pb2.QuaProgram.RampToZeroStatement(
        loc=loc,
        qe=inc_qua_pb2.QuaProgram.QuantumElementReference(name=element, loc=loc),
        duration=duration_val,
    )
    scopes_manager.append_statement(inc_qua_pb2.QuaProgram.AnyStatement(rampToZero=statement))


def load_waveform(
    pulse: str,
    waveform_index: Scalar[int],
    element: str,
) -> None:
    """
    Load a waveform from the waveform array to be used in a pulse.

    Args:
        pulse (str): The name of an `operation` to be performed, as defined in the element in the quantum machine
            configuration.
        waveform_index (int, QUA int): The index of the waveform to be loaded from the waveform_array that was
            configured for the pulse.
        element (str): The name of the element, as defined in the quantum machine configuration.
    """
    loc = _get_loc()
    scopes_manager.program_scope.add_used_capability(QopCaps.waveform_array)
    statement = inc_qua_pb2.QuaProgram.AnyStatement(
        loadWaveform=inc_qua_pb2.QuaProgram.LoadWaveformStatement(
            loc=loc,
            pulse=inc_qua_pb2.QuaProgram.PulseReference(name=pulse, loc=loc),
            qe=inc_qua_pb2.QuaProgram.QuantumElementReference(name=element, loc=loc),
            waveform_index=to_scalar_pb_expression(waveform_index),
        )
    )
    scopes_manager.append_statement(statement)



from typing import TYPE_CHECKING, Optional, Sequence, cast

from qm.grpc.qm.pb import inc_qua_pb2
from qm.exceptions import QmQuaException, NoScopeFoundException
from qm.qua._scope_management._core_scopes import _BaseScope, _ProgramScope, _get_scopes_stack

if TYPE_CHECKING:
    from qm.qua._dsl.stream_processing.stream_processing import _OutputStream


class _ScopesStackManager:
    """
    Provides higher-level access to the scopes stack, used by DSL functions and, optionally, by the scope classes.
    We do not want the scopes stack to be exposed to the DSL functions, so we provide this manager.
    """

    def __init__(self) -> None:
        self._port_condition: Optional[inc_qua_pb2.QuaProgram.AnyScalarExpression] = None

    @property
    def program_scope(self) -> _ProgramScope:
        scopes_stack = _get_scopes_stack()
        if len(scopes_stack) == 0:
            raise NoScopeFoundException("No program scope found")
        return cast(_ProgramScope, scopes_stack[0])

    @property
    def current_scope(self) -> _BaseScope:
        scopes_stack = _get_scopes_stack()
        if len(scopes_stack) == 0:
            raise NoScopeFoundException("No scope found")
        return scopes_stack[-1]

    @property
    def scope_stack(self) -> Sequence[_BaseScope]:
        return tuple(_get_scopes_stack())

    def append_statement(self, statement: inc_qua_pb2.QuaProgram.AnyStatement) -> None:
        self.current_scope.append_statement(statement)

    def append_output_stream(self, stream: "_OutputStream") -> None:
        self.program_scope.result_analysis_scope.append_output_stream(stream)

    # The function is private since it should only be used by the PortConditionScope, and should not be used by the DSL functions.
    def _set_port_condition(self, condition: inc_qua_pb2.QuaProgram.AnyScalarExpression) -> None:
        scopes_stack = _get_scopes_stack()
        if len(scopes_stack) == 0:
            raise NoScopeFoundException("No program scope found")
        if self._port_condition is not None:
            raise QmQuaException("port_condition already set")
        self._port_condition = condition

    # The function is private since it should only be used by the PortConditionScope, and should not be used by the DSL functions.
    def _unset_port_condition(self) -> None:
        self._port_condition = None

    @property
    def port_condition(self) -> Optional[inc_qua_pb2.QuaProgram.AnyScalarExpression]:
        return self._port_condition


scopes_manager = _ScopesStackManager()




# Wrapper to enforce deterministic serialization
def deterministic_serialize_wrapper(orig_serialize_func):
    def wrapped(msg):
        return orig_serialize_func(msg, deterministic=True)
    return wrapped

# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from qm.grpc.qm.grpc.v2 import qmm_api_pb2 as qm_dot_grpc_dot_v2_dot_qmm__api__pb2

GRPC_GENERATED_VERSION = '1.68.1'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in qm/grpc/v2/qmm_api_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class QmmServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetVersion = channel.unary_unary(
                '/qm.grpc.v2.QmmService/GetVersion',
                request_serializer=deterministic_serialize_wrapper(qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetVersionRequest.SerializeToString),
                response_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetVersionResponse.FromString,
                _registered_method=True)
        self.HealthCheck = channel.unary_unary(
                '/qm.grpc.v2.QmmService/HealthCheck',
                request_serializer=deterministic_serialize_wrapper(qm_dot_grpc_dot_v2_dot_qmm__api__pb2.HealthCheckRequest.SerializeToString),
                response_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.HealthCheckResponse.FromString,
                _registered_method=True)
        self.GetControllers = channel.unary_unary(
                '/qm.grpc.v2.QmmService/GetControllers',
                request_serializer=deterministic_serialize_wrapper(qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetControllersRequest.SerializeToString),
                response_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetControllersResponse.FromString,
                _registered_method=True)
        self.OpenQuantumMachine = channel.unary_unary(
                '/qm.grpc.v2.QmmService/OpenQuantumMachine',
                request_serializer=deterministic_serialize_wrapper(qm_dot_grpc_dot_v2_dot_qmm__api__pb2.OpenQuantumMachineRequest.SerializeToString),
                response_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.OpenQuantumMachineResponse.FromString,
                _registered_method=True)
        self.GetQuantumMachine = channel.unary_unary(
                '/qm.grpc.v2.QmmService/GetQuantumMachine',
                request_serializer=deterministic_serialize_wrapper(qm_dot_grpc_dot_v2_dot_qmm__api__pb2.QmmServiceGetQuantumMachineRequest.SerializeToString),
                response_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetQuantumMachineResponse.FromString,
                _registered_method=True)
        self.ListOpenQuantumMachines = channel.unary_unary(
                '/qm.grpc.v2.QmmService/ListOpenQuantumMachines',
                request_serializer=deterministic_serialize_wrapper(qm_dot_grpc_dot_v2_dot_qmm__api__pb2.ListOpenQuantumMachinesRequest.SerializeToString),
                response_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.ListOpenQuantumMachinesResponse.FromString,
                _registered_method=True)
        self.CloseAllQuantumMachines = channel.unary_unary(
                '/qm.grpc.v2.QmmService/CloseAllQuantumMachines',
                request_serializer=deterministic_serialize_wrapper(qm_dot_grpc_dot_v2_dot_qmm__api__pb2.CloseAllQuantumMachinesRequest.SerializeToString),
                response_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.CloseAllQuantumMachinesResponse.FromString,
                _registered_method=True)
        self.ClearAllJobResults = channel.unary_unary(
                '/qm.grpc.v2.QmmService/ClearAllJobResults',
                request_serializer=deterministic_serialize_wrapper(qm_dot_grpc_dot_v2_dot_qmm__api__pb2.ClearAllJobResultsRequest.SerializeToString),
                response_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.ClearAllJobResultsResponse.FromString,
                _registered_method=True)
        self.GetJobs = channel.unary_unary(
                '/qm.grpc.v2.QmmService/GetJobs',
                request_serializer=deterministic_serialize_wrapper(qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetJobsRequest.SerializeToString),
                response_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetJobsResponse.FromString,
                _registered_method=True)
        self.ResetDataProcessing = channel.unary_unary(
                '/qm.grpc.v2.QmmService/ResetDataProcessing',
                request_serializer=deterministic_serialize_wrapper(qm_dot_grpc_dot_v2_dot_qmm__api__pb2.QmmServiceResetDataProcessingRequest.SerializeToString),
                response_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.ResetDataProcessingResponse.FromString,
                _registered_method=True)
        self.Simulate = channel.unary_unary(
                '/qm.grpc.v2.QmmService/Simulate',
                request_serializer=deterministic_serialize_wrapper(qm_dot_grpc_dot_v2_dot_qmm__api__pb2.QmmServiceSimulateRequest.SerializeToString),
                response_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.QmmServiceSimulateResponse.FromString,
                _registered_method=True)


class QmmServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetVersion(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def HealthCheck(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetControllers(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def OpenQuantumMachine(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetQuantumMachine(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListOpenQuantumMachines(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CloseAllQuantumMachines(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ClearAllJobResults(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetJobs(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ResetDataProcessing(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Simulate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_QmmServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetVersion': grpc.unary_unary_rpc_method_handler(
                    servicer.GetVersion,
                    request_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetVersionRequest.FromString,
                    response_serializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetVersionResponse.SerializeToString,
            ),
            'HealthCheck': grpc.unary_unary_rpc_method_handler(
                    servicer.HealthCheck,
                    request_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.HealthCheckRequest.FromString,
                    response_serializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.HealthCheckResponse.SerializeToString,
            ),
            'GetControllers': grpc.unary_unary_rpc_method_handler(
                    servicer.GetControllers,
                    request_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetControllersRequest.FromString,
                    response_serializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetControllersResponse.SerializeToString,
            ),
            'OpenQuantumMachine': grpc.unary_unary_rpc_method_handler(
                    servicer.OpenQuantumMachine,
                    request_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.OpenQuantumMachineRequest.FromString,
                    response_serializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.OpenQuantumMachineResponse.SerializeToString,
            ),
            'GetQuantumMachine': grpc.unary_unary_rpc_method_handler(
                    servicer.GetQuantumMachine,
                    request_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.QmmServiceGetQuantumMachineRequest.FromString,
                    response_serializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetQuantumMachineResponse.SerializeToString,
            ),
            'ListOpenQuantumMachines': grpc.unary_unary_rpc_method_handler(
                    servicer.ListOpenQuantumMachines,
                    request_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.ListOpenQuantumMachinesRequest.FromString,
                    response_serializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.ListOpenQuantumMachinesResponse.SerializeToString,
            ),
            'CloseAllQuantumMachines': grpc.unary_unary_rpc_method_handler(
                    servicer.CloseAllQuantumMachines,
                    request_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.CloseAllQuantumMachinesRequest.FromString,
                    response_serializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.CloseAllQuantumMachinesResponse.SerializeToString,
            ),
            'ClearAllJobResults': grpc.unary_unary_rpc_method_handler(
                    servicer.ClearAllJobResults,
                    request_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.ClearAllJobResultsRequest.FromString,
                    response_serializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.ClearAllJobResultsResponse.SerializeToString,
            ),
            'GetJobs': grpc.unary_unary_rpc_method_handler(
                    servicer.GetJobs,
                    request_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetJobsRequest.FromString,
                    response_serializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetJobsResponse.SerializeToString,
            ),
            'ResetDataProcessing': grpc.unary_unary_rpc_method_handler(
                    servicer.ResetDataProcessing,
                    request_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.QmmServiceResetDataProcessingRequest.FromString,
                    response_serializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.ResetDataProcessingResponse.SerializeToString,
            ),
            'Simulate': grpc.unary_unary_rpc_method_handler(
                    servicer.Simulate,
                    request_deserializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.QmmServiceSimulateRequest.FromString,
                    response_serializer=qm_dot_grpc_dot_v2_dot_qmm__api__pb2.QmmServiceSimulateResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'qm.grpc.v2.QmmService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('qm.grpc.v2.QmmService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class QmmService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetVersion(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qm.grpc.v2.QmmService/GetVersion',
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetVersionRequest.SerializeToString,
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetVersionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def HealthCheck(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qm.grpc.v2.QmmService/HealthCheck',
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.HealthCheckRequest.SerializeToString,
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.HealthCheckResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetControllers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qm.grpc.v2.QmmService/GetControllers',
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetControllersRequest.SerializeToString,
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetControllersResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def OpenQuantumMachine(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qm.grpc.v2.QmmService/OpenQuantumMachine',
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.OpenQuantumMachineRequest.SerializeToString,
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.OpenQuantumMachineResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetQuantumMachine(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qm.grpc.v2.QmmService/GetQuantumMachine',
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.QmmServiceGetQuantumMachineRequest.SerializeToString,
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetQuantumMachineResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ListOpenQuantumMachines(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qm.grpc.v2.QmmService/ListOpenQuantumMachines',
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.ListOpenQuantumMachinesRequest.SerializeToString,
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.ListOpenQuantumMachinesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CloseAllQuantumMachines(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qm.grpc.v2.QmmService/CloseAllQuantumMachines',
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.CloseAllQuantumMachinesRequest.SerializeToString,
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.CloseAllQuantumMachinesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ClearAllJobResults(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qm.grpc.v2.QmmService/ClearAllJobResults',
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.ClearAllJobResultsRequest.SerializeToString,
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.ClearAllJobResultsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetJobs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qm.grpc.v2.QmmService/GetJobs',
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetJobsRequest.SerializeToString,
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.GetJobsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ResetDataProcessing(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qm.grpc.v2.QmmService/ResetDataProcessing',
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.QmmServiceResetDataProcessingRequest.SerializeToString,
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.ResetDataProcessingResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Simulate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/qm.grpc.v2.QmmService/Simulate',
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.QmmServiceSimulateRequest.SerializeToString,
            qm_dot_grpc_dot_v2_dot_qmm__api__pb2.QmmServiceSimulateResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



from enum import Enum
from typing import Dict, Union

from qm.grpc.octave.v1 import api_pb2


class RFInputRFSource(Enum):
    """ """

    Loopback_RF_out_1 = 1
    Loopback_RF_out_2 = 2
    Loopback_RF_out_3 = 3
    Loopback_RF_out_4 = 4
    Loopback_RF_out_5 = 5
    RF_in = 6
    Off = 7


synth_slot_index_to_panel_mapping: Dict[int, str] = {
    1: "5",
    2: "6",
    3: "1",
    4: "calibration",
    5: "3",
    6: "2",
}


def slot_index_to_panel_mapping(index: int, module_type: api_pb2.OctaveModule) -> Union[int, str]:
    if module_type == api_pb2.OctaveModule.OCTAVE_MODULE_SYNTHESIZER:
        return synth_slot_index_to_panel_mapping[index]

    return index


octave_module_to_module_name_mapping: Dict[api_pb2.OctaveModule.ValueType, str] = {  # type: ignore[name-defined]
    api_pb2.OctaveModule.OCTAVE_MODULE_SYNTHESIZER: "Synthesizer",
    api_pb2.OctaveModule.OCTAVE_MODULE_IF_DOWNCONVERTER: "IF",
    api_pb2.OctaveModule.OCTAVE_MODULE_RF_DOWNCONVERTER: "RF Down-converter",
    api_pb2.OctaveModule.OCTAVE_MODULE_RF_UPCONVERTER: "RF Up-converter",
    api_pb2.OctaveModule.OCTAVE_MODULE_SOM: "System",
    api_pb2.OctaveModule.OCTAVE_MODULE_MOTHERBOARD: "Motherboard",
}



import json
import logging
from typing import Any, Dict, Tuple, Union, TypeVar, Callable, Optional, cast

import grpc
from multidict import MultiDict
from google.protobuf.message import Message
from google.protobuf.json_format import MessageToJson

from qm.exceptions import QmQuaException
from qm.api.models.debug_data import DebugData

# gRPC typing helpers
ClientCallDetails = grpc.ClientCallDetails
ReqT = TypeVar("ReqT")
RespT = TypeVar("RespT")


# Runtime-safe fallback — no subscripting
UnaryUnaryResult = grpc.Future
UnaryStreamResult = grpc.Call

# Important
# due to mypy-runtime grpc discrepancies I place some mypy ignores in this file after seeing it will be very hard to support both mypy and runtime
UnaryUnaryContinuation = Callable[..., UnaryUnaryResult]  # type: ignore[type-arg]
UnaryStreamContinuation = Callable[..., UnaryStreamResult]

logger = logging.getLogger(__name__)

MetadataItem = Tuple[str, Union[str, bytes]]
Metadata = Optional[Tuple[MetadataItem, ...]]


class _ClientCallDetails(grpc.ClientCallDetails):
    def __init__(
        self,
        method: str,
        timeout: Optional[float],
        metadata: Metadata,
        credentials: Optional[Any] = None,
        wait_for_ready: Optional[bool] = None,
        compression: Optional[Any] = None,
    ):
        self.method = method
        self.timeout = timeout
        self.metadata = metadata
        self.credentials = credentials
        self.wait_for_ready = wait_for_ready
        self.compression = compression


class DebugInterceptor(
    grpc.UnaryUnaryClientInterceptor,
    grpc.UnaryStreamClientInterceptor,
):
    def __init__(self, debug_data: DebugData) -> None:
        self.debug_data = debug_data

    def intercept_unary_unary(  # type: ignore[override]
        self,
        continuation: UnaryUnaryContinuation,
        client_call_details: ClientCallDetails,
        request: ReqT,
    ) -> UnaryUnaryResult:  # type: ignore[type-arg]
        self._log_request(client_call_details, request)
        future = continuation(client_call_details, request)
        self._log_response_metadata(future, False)
        return future

    def intercept_unary_stream(  # type: ignore[override]
        self,
        continuation: UnaryStreamContinuation,
        client_call_details: ClientCallDetails,
        request: ReqT,
    ) -> UnaryStreamResult:
        self._log_request(client_call_details, request)
        response = continuation(client_call_details, request)
        self._log_response_metadata(response, True)
        return response

    def _log_request(
        self,
        client_call_details: ClientCallDetails,
        request: ReqT,
    ) -> None:
        logger.debug("-----------request start-----------")
        logger.debug("   ---    request headers    ---   ")
        logger.debug(f"method:        {client_call_details.method}")

        if client_call_details.metadata:
            metadata_dict = dict(client_call_details.metadata)
            logger.debug(f"metadata:     {json.dumps(metadata_dict, indent=4)}")
        else:
            logger.debug("metadata:     None")

        if client_call_details.timeout is not None:
            logger.debug(f"timeout:      {client_call_details.timeout}")
        else:
            logger.debug("timeout:      None")

        if isinstance(request, Message):
            logger.debug("   ---    request message    ---   ")
            try:
                message_json = MessageToJson(request, indent=4)
                logger.debug(f"message:      {message_json}")
            except Exception as e:
                logger.debug(f"message:      [Could not serialize: {e}]")

            logger.debug("------------end request------------")
        else:
            raise QmQuaException(f"got request of type {type(request)} instead of protobuf Message")

    def _log_response_metadata(
        self,
        response: Union[UnaryUnaryResult, UnaryStreamResult],  # type: ignore[type-arg]
        is_stream: bool,
    ) -> None:
        try:
            initial_metadata = response.initial_metadata()  # type: ignore[union-attr]
            metadata_dict = dict(initial_metadata)
            logger.debug(f"Collected response metadata: {json.dumps(metadata_dict, indent=4)}")
            self.debug_data.append(cast(MultiDict[Union[str, bytes]], metadata_dict))
        except Exception as e:
            logger.debug(f"Could not extract response metadata: {e}")


class AddHeadersInterceptor(
    grpc.UnaryUnaryClientInterceptor,
    grpc.UnaryStreamClientInterceptor,
):
    def __init__(self, headers: Dict[str, str]):
        self.headers = headers

    def intercept_unary_unary(  # type: ignore[override]
        self,
        continuation: UnaryUnaryContinuation,
        client_call_details: ClientCallDetails,
        request: ReqT,
    ) -> UnaryUnaryResult:  # type: ignore[type-arg]
        new_details = self._add_headers_to_call_details(client_call_details)
        return continuation(new_details, request)

    def intercept_unary_stream(  # type: ignore[override]
        self,
        continuation: UnaryStreamContinuation,
        client_call_details: ClientCallDetails,
        request: ReqT,
    ) -> UnaryStreamResult:
        new_details = self._add_headers_to_call_details(client_call_details)
        return continuation(new_details, request)

    def _add_headers_to_call_details(self, client_call_details: ClientCallDetails) -> ClientCallDetails:
        new_metadata = list(self.headers.items())

        if client_call_details.metadata:
            new_metadata.extend((k, v.decode() if isinstance(v, bytes) else v) for k, v in client_call_details.metadata)

        return _ClientCallDetails(
            method=client_call_details.method,
            timeout=client_call_details.timeout,
            metadata=tuple(new_metadata),
            credentials=getattr(client_call_details, "credentials", None),
            wait_for_ready=getattr(client_call_details, "wait_for_ready", None),
            compression=getattr(client_call_details, "compression", None),
        )



import dataclasses
from typing_extensions import dataclass_transform
from typing import (
    Any,
    List,
    Type,
    Union,
    Generic,
    Literal,
    Mapping,
    TypeVar,
    Callable,
    Protocol,
    get_args,
    overload,
    get_origin,
    get_type_hints,
)

from qm.type_hinting import NumberT
from qm.grpc.qm.pb import inc_qua_pb2
from qm.type_hinting.general import DataclassProtocol
from qm.qua._dsl.variable_handling import _declare_struct_array_variable
from qm.exceptions import ReservedFieldNameError, InvalidQuaArraySubclassError
from qm.qua._expressions import NSize, QuaStructReference, QuaStructArrayVariable

# A more user-friendly alias for QuaStructArrayVariable.
# In the future, we might want to use QuaArray for other types beyond QuaStructArrayVariable.
# For now, it's simply set to QuaStructArrayVariable since that's the only current use case.
QuaArray = QuaStructArrayVariable


class StructArrayFactory(Generic[NumberT]):
    def __init__(self, array_type: Type[NumberT], size: int, position: int):
        self._array_type: Type[NumberT] = array_type
        self._size = size
        self._position = position

    def create(self, struct_ref: QuaStructReference) -> QuaStructArrayVariable[NumberT, NSize]:
        return _declare_struct_array_variable(
            t=self._array_type, size=self._size, position=self._position, struct_ref=struct_ref
        )

    @property
    def underlying_declaration(self) -> inc_qua_pb2.QuaProgram.VarDeclaration:
        # This function is called before 'create', so the real struct member declaration isn't available yet.
        # We do not really need a real declaration, only the 'metadata' of the struct array - type, size and position in struct.
        return QuaStructArrayVariable(
            "", self._array_type, self._size, self._position, QuaStructReference("")
        ).declaration_statement


class _QuaStruct(DataclassProtocol, Protocol):
    __members_initializers__: Mapping[str, StructArrayFactory[Any]]
    __underlying_declarations__: List[inc_qua_pb2.QuaProgram.VarDeclaration]
    struct_reference: QuaStructReference


_T = TypeVar("_T")


@dataclass_transform()
@overload
def qua_struct(_cls: Union[Type[_T], None]) -> Type[_T]:
    """
    Decorate a class as a QUA struct.

    Args:
        _cls (type[_T] | None): The class to decorate.
    """
    ...


@dataclass_transform()
@overload
def qua_struct() -> Callable[[Type[_T]], Type[_T]]:
    """
    Return a decorator that marks a class as a QUA struct.

    This form is equivalent to using ``@qua_struct`` directly.
    """
    ...


@dataclass_transform()
def qua_struct(_cls: Union[Type[_T], None] = None) -> Union[Callable[[Type[_T]], Type[_T]], Type[_T]]:
    """
    Decorate a class as a QUA struct.

    A QUA struct defines the packet schema used by OPNIC input and output streams.
    Each annotated field becomes a QUA array member in the packet, and every field must use
    the form ``QuaArray[element_type, size]``. Even a single scalar packet value must be declared
    as a size-1 array, for example ``QuaArray[int, 1]``.

    After decorating the class, create QUA struct instances inside a QUA program with
    [qm.qua.declare_struct][]. Use those instances with [qm.qua.receive_from_stream][] and
    [qm.qua.send_to_stream][].

    See [OPNIC Hybrid Link](../../Guides/OPNIC_Guide.md) for end-to-end examples.

    Example:
        ```python
        @qua_struct
        class Packet:
            data: QuaArray[int, 1]
            flags: QuaArray[bool, 2]
        ```

    Args:
        _cls (type[_T] | None): The class to decorate. If omitted, returns a decorator.
    """

    def validate_field_type(annotation: Any) -> None:
        field_type = get_origin(annotation)
        # Check whether the field type is a class, and if it is specifically a subclass of QuaArray
        if isinstance(field_type, type) and issubclass(field_type, QuaArray):
            if len(get_args(annotation)) != 2:
                raise InvalidQuaArraySubclassError(
                    f"Invalid QuaArray subclass: {field_type}. "
                    "Expected exactly two type arguments (e.g., QuaArray[int, 5]) "
                    "representing the element type and array size."
                )
        else:
            # The field might not be annotated with generics, causing get_origin to return None.
            if field_type is None:
                field_type = annotation
            raise TypeError(f"Type '{field_type.__name__}' is not supported as a QUA struct field type")

    def get_members_initializers(cls: Type[_T]) -> Mapping[str, StructArrayFactory[Any]]:
        output = {}
        for index, (name, annotation) in enumerate(get_type_hints(cls).items()):
            if name in get_type_hints(_QuaStruct).keys():
                raise ReservedFieldNameError(f"Field name '{name}' is reserved for internal use within a QUA struct")

            validate_field_type(annotation)

            array_type, array_size = get_args(annotation)
            array_size_type = get_origin(array_size)
            if array_size_type == Literal:  # For "proper" type hinting
                array_size = get_args(array_size)[0]

            output[name] = StructArrayFactory(array_type, array_size, index)

        return output

    def add_reference_field(cls: Type[_T]) -> None:
        # Add the `_struct_reference` field to the class
        cls.__annotations__["_struct_reference"] = QuaStructReference  # noqa
        setattr(cls, "_struct_reference", dataclasses.field())  # noqa: B010

        # Add a read-only property for 'reference'
        @property  # type: ignore
        def struct_reference(self) -> QuaStructReference:  # type: ignore[no-untyped-def]
            return self._struct_reference  # type: ignore

        cls.struct_reference = struct_reference  # type: ignore[attr-defined]

    def create_qua_struct(cls: Type[_T]) -> Type[_T]:
        members_initializers = get_members_initializers(cls)
        underlying_declarations = [
            member_factory.underlying_declaration for member_factory in members_initializers.values()
        ]

        add_reference_field(cls)
        cls.__members_initializers__ = members_initializers  # type: ignore[attr-defined]
        cls.__underlying_declarations__ = underlying_declarations  # type: ignore[attr-defined]

        dataclass_cls = dataclasses.dataclass(cls)
        return dataclass_cls

    return create_qua_struct(_cls) if _cls is not None else create_qua_struct



from qm.waveform_report._type_hints import WaveformReportType
from qm.waveform_report._waveform_report import (
    Event,
    AdcAcquisition,
    WaveformReport,
    PlayedAnalogWaveform,
    PlayedDigitalWaveform,
    PlayedMwAnalogWaveform,
)

__all__ = [
    "WaveformReport",
    "WaveformReportType",
    "PlayedAnalogWaveform",
    "PlayedDigitalWaveform",
    "PlayedMwAnalogWaveform",
    "AdcAcquisition",
    "Event",
]



import abc
from typing import Any, Dict, Type, Union, Optional, Sequence, overload

from qm.exceptions import QmQuaException
from qm.qua._expressions import QuaVariable, QuaArrayCell, QuaArrayVariable
from qm.qua._dsl.measure.analog_measure_process import RawTimeTagging as AnalogRawTimeTagging
from qm.qua._dsl.measure.digital_measure_process import RawTimeTagging as DigitalRawTimeTagging
from qm.qua._dsl.measure.digital_measure_process import Counting as DigitalMeasureProcessCounting
from qm.qua._dsl.measure.analog_measure_process import (
    BareIntegration,
    BasicIntegration,
    DemodIntegration,
    DualMeasureProcess,
    HighResTimeTagging,
    ScalarProcessTarget,
    VectorProcessTarget,
    DualDemodIntegration,
    SlicedAnalogTimeDivision,
    AccumulatedAnalogTimeDivision,
    MovingWindowAnalogTimeDivision,
)

DEFAULT_OUT1 = "out1"
DEFAULT_OUT2 = "out2"


class _AccumulationMethod(metaclass=abc.ABCMeta):
    loc = ""

    @classmethod
    def _full_target(cls, target: Union[QuaVariable[float], QuaArrayCell[float]]) -> ScalarProcessTarget:
        return ScalarProcessTarget(cls.loc, target)

    @classmethod
    def _sliced_target(cls, target: QuaArrayVariable[float], samples_per_chunk: int) -> VectorProcessTarget:
        analog_time_division = SlicedAnalogTimeDivision(cls.loc, samples_per_chunk)
        return VectorProcessTarget(cls.loc, target, analog_time_division)

    @classmethod
    def _accumulated_target(cls, target: QuaArrayVariable[float], samples_per_chunk: int) -> VectorProcessTarget:
        analog_time_division = AccumulatedAnalogTimeDivision(cls.loc, samples_per_chunk)
        return VectorProcessTarget(cls.loc, target, analog_time_division)

    @classmethod
    def _moving_window_target(
        cls, target: QuaArrayVariable[float], samples_per_chunk: int, chunks_per_window: int
    ) -> VectorProcessTarget:
        analog_time_division = MovingWindowAnalogTimeDivision(cls.loc, samples_per_chunk, chunks_per_window)
        return VectorProcessTarget(cls.loc, target, analog_time_division)


class _RealAccumulationMethod(_AccumulationMethod, metaclass=abc.ABCMeta):
    @classmethod
    @abc.abstractmethod
    def _get_return_func(cls) -> Type[BasicIntegration]:
        pass

    @classmethod
    def full(
        cls, iw: str, target: Union[QuaVariable[float], QuaArrayCell[float]], element_output: str = ""
    ) -> BasicIntegration:
        return cls._get_return_func()(element_output, iw, cls._full_target(target))

    @classmethod
    def sliced(
        cls,
        iw: str,
        target: QuaArrayVariable[float],
        samples_per_chunk: int,
        element_output: str = "",
    ) -> BasicIntegration:
        return cls._get_return_func()(element_output, iw, cls._sliced_target(target, samples_per_chunk))

    @classmethod
    def accumulated(
        cls,
        iw: str,
        target: QuaArrayVariable[float],
        samples_per_chunk: int,
        element_output: str = "",
    ) -> BasicIntegration:
        return cls._get_return_func()(element_output, iw, cls._accumulated_target(target, samples_per_chunk))

    @classmethod
    def moving_window(
        cls,
        iw: str,
        target: QuaArrayVariable[float],
        samples_per_chunk: int,
        chunks_per_window: int,
        element_output: str = "",
    ) -> BasicIntegration:
        return cls._get_return_func()(
            element_output,
            iw,
            cls._moving_window_target(target, samples_per_chunk, chunks_per_window),
        )


class _DualAccumulationMethod(_AccumulationMethod, metaclass=abc.ABCMeta):
    @classmethod
    @abc.abstractmethod
    def _get_return_func(cls) -> Type[DualMeasureProcess]:
        pass

    @staticmethod
    def _make_dict_from_args(args: Sequence[object], names: Sequence[str]) -> Dict[str, object]:
        return {name: arg for name, arg in zip(names, args)}

    @overload
    @classmethod
    def full(
        cls,
        iw1: str,
        iw2: str,
        target: Union[QuaVariable[float], QuaArrayCell[float]],
    ) -> DualMeasureProcess:
        ...

    @overload
    @classmethod
    def full(
        cls,
        iw1: str,
        element_output1: str,
        iw2: str,
        element_output2: str,
        target: Union[QuaVariable[float], QuaArrayCell[float]],
    ) -> DualMeasureProcess:
        ...

    @classmethod
    def full(cls, *args: Any, **kwargs: Any) -> DualMeasureProcess:
        if len(args) + len(kwargs) == 3:
            kwargs.update(_DualAccumulationMethod._make_dict_from_args(args, ["iw1", "iw2", "target"]))
            kwargs["element_output1"] = DEFAULT_OUT1
            kwargs["element_output2"] = DEFAULT_OUT2
            return cls.full(**kwargs)
        elif len(args) + len(kwargs) == 5:
            kwargs.update(
                _DualAccumulationMethod._make_dict_from_args(
                    args, ["iw1", "element_output1", "iw2", "element_output2", "target"]
                )
            )
        else:
            raise QmQuaException("Invalid number of arguments")

        return cls._get_return_func()(
            kwargs["element_output1"],
            kwargs["element_output2"],
            kwargs["iw1"],
            kwargs["iw2"],
            cls._full_target(kwargs["target"]),
        )

    @overload
    @classmethod
    def sliced(
        cls,
        iw1: str,
        iw2: str,
        samples_per_chunk: int,
        target: QuaArrayVariable[float],
    ) -> DualMeasureProcess:
        pass

    @overload
    @classmethod
    def sliced(
        cls,
        iw1: str,
        element_output1: str,
        iw2: str,
        element_output2: str,
        samples_per_chunk: int,
        target: QuaArrayVariable[float],
    ) -> DualMeasureProcess:
        pass

    @classmethod
    def sliced(cls, *args: Any, **kwargs: Any) -> DualMeasureProcess:
        """This feature is currently not supported in QUA"""
        if len(args) + len(kwargs) == 4:
            kwargs.update(
                _DualAccumulationMethod._make_dict_from_args(args, ["iw1", "iw2", "samples_per_chunk", "target"])
            )
            kwargs["element_output1"] = DEFAULT_OUT1
            kwargs["element_output2"] = DEFAULT_OUT2
            return cls.sliced(**kwargs)
        elif len(args) + len(kwargs) == 6:
            kwargs.update(
                _DualAccumulationMethod._make_dict_from_args(
                    args,
                    ["iw1", "element_output1", "iw2", "element_output2", "samples_per_chunk", "target"],
                )
            )
        else:
            raise QmQuaException("Invalid number of arguments")

        return cls._get_return_func()(
            kwargs["element_output1"],
            kwargs["element_output2"],
            kwargs["iw1"],
            kwargs["iw2"],
            cls._sliced_target(kwargs["target"], kwargs["samples_per_chunk"]),
        )

    @overload
    @classmethod
    def accumulated(
        cls,
        iw1: str,
        iw2: str,
        samples_per_chunk: int,
        target: QuaArrayVariable[float],
    ) -> DualMeasureProcess:
        pass

    @overload
    @classmethod
    def accumulated(
        cls,
        iw1: str,
        element_output1: str,
        iw2: str,
        element_output2: str,
        samples_per_chunk: int,
        target: QuaArrayVariable[float],
    ) -> DualMeasureProcess:
        pass

    @classmethod
    def accumulated(cls, *args: Any, **kwargs: Any) -> DualMeasureProcess:
        """This feature is currently not supported in QUA"""
        if len(args) + len(kwargs) == 4:
            kwargs.update(
                _DualAccumulationMethod._make_dict_from_args(args, ["iw1", "iw2", "samples_per_chunk", "target"])
            )
            kwargs["element_output1"] = DEFAULT_OUT1
            kwargs["element_output2"] = DEFAULT_OUT2
            return cls.accumulated(**kwargs)
        elif len(args) + len(kwargs) == 6:
            kwargs.update(
                _DualAccumulationMethod._make_dict_from_args(
                    args,
                    ["iw1", "element_output1", "iw2", "element_output2", "samples_per_chunk", "target"],
                )
            )
        else:
            raise QmQuaException("Invalid number of arguments")

        return cls._get_return_func()(
            kwargs["element_output1"],
            kwargs["element_output2"],
            kwargs["iw1"],
            kwargs["iw2"],
            cls._accumulated_target(kwargs["target"], kwargs["samples_per_chunk"]),
        )

    @overload
    @classmethod
    def moving_window(
        cls,
        iw1: str,
        iw2: str,
        samples_per_chunk: int,
        chunks_per_window: int,
        target: QuaArrayVariable[float],
    ) -> DualMeasureProcess:
        pass

    @overload
    @classmethod
    def moving_window(
        cls,
        iw1: str,
        element_output1: str,
        iw2: str,
        element_output2: str,
        samples_per_chunk: int,
        chunks_per_window: int,
        target: QuaArrayVariable[float],
    ) -> DualMeasureProcess:
        pass

    @classmethod
    def moving_window(cls, *args: Any, **kwargs: Any) -> DualMeasureProcess:
        """This feature is currently not supported in QUA"""
        if len(args) + len(kwargs) == 5:
            kwargs.update(
                _DualAccumulationMethod._make_dict_from_args(
                    args, ["iw1", "iw2", "samples_per_chunk", "chunks_per_window", "target"]
                )
            )
            kwargs["element_output1"] = DEFAULT_OUT1
            kwargs["element_output2"] = DEFAULT_OUT2
            return cls.moving_window(**kwargs)
        elif len(args) + len(kwargs) == 7:
            kwargs.update(
                _DualAccumulationMethod._make_dict_from_args(
                    args,
                    [
                        "iw1",
                        "element_output1",
                        "iw2",
                        "element_output2",
                        "samples_per_chunk",
                        "chunks_per_window",
                        "target",
                    ],
                )
            )
        else:
            raise QmQuaException("Invalid number of arguments")

        return cls._get_return_func()(
            kwargs["element_output1"],
            kwargs["element_output2"],
            kwargs["iw1"],
            kwargs["iw2"],
            cls._moving_window_target(kwargs["target"], kwargs["samples_per_chunk"], kwargs["chunks_per_window"]),
        )


class time_tagging:
    """A base class for specifying the time tagging process in the [measure][qm.qua.measure] statement.
    These are the options that can be used inside the measure command as part of the ``time_tagging`` process.
    """

    loc = ""

    @staticmethod
    def analog(
        target: QuaArrayVariable[int],
        max_time: int,
        targetLen: Optional[QuaVariable[int]] = None,
        element_output: str = "",
    ) -> AnalogRawTimeTagging:
        """Performs time tagging. See [Time tagging](../../Guides/features.md#time-tagging).

        Args:
            target (QUA array of type int): The QUA array into which the
                times of the detected pulses are saved (the units depend on the system, see the [documentation](../../Guides/features.md#time-tagging))
            max_time (int): The time in which pulses are detected
                (Must be larger than the pulse duration)
            targetLen (QUA int): A QUA int which will get the number of
                pulses detected
            element_output (str): The output of an element from which to get the pulses.
                Required when there are multiple outputs in the element. Optional otherwise.
        """
        return AnalogRawTimeTagging(element_output, target, targetLen, max_time)

    @staticmethod
    def digital(
        target: QuaArrayVariable[int],
        max_time: int,
        targetLen: Optional[QuaVariable[int]] = None,
        element_output: str = "",
    ) -> DigitalRawTimeTagging:
        """Performs time tagging from the attached OPD.
         See [Time tagging](../../Guides/features.md#time-tagging).

        -- Available with the OPD addon --

        Args:
            target (QUA array of type int): The QUA array into which the
                times of the detected pulses are saved (in ns)
            max_time (int): The time in which pulses are detected
                (Must be larger than the pulse duration)
            targetLen (QUA int): A QUA int which will get the number of
                pulses detected
            element_output (str): The output of an element from which to get the pulses.
                Required when there are multiple outputs in the element. Optional otherwise.
        """
        return DigitalRawTimeTagging(element_output, target, targetLen, max_time)

    @staticmethod
    def high_res(
        target: QuaArrayVariable[int],
        max_time: int,
        targetLen: Optional[QuaVariable[int]] = None,
        element_output: str = "",
    ) -> HighResTimeTagging:
        """Performs high resolution time tagging. See [Time tagging](../../Guides/features.md#time-tagging).

        -- Available from QOP 2.0 --

        Args:
            target (QUA array of type int): The QUA array into which the
                times of the detected pulses are saved (in ps)
            max_time (int): The time in which pulses are detected
                (Must be larger than the pulse duration)
            targetLen (QUA int): A QUA int which will get the number of
                pulses detected
            element_output (str): The output of an element from which to get the pulses.
                Required when there are multiple outputs in the element. Optional otherwise.
        """
        return HighResTimeTagging(element_output, target, targetLen, max_time)


class counting:
    """A base class for specifying the counting process in the [measure][qm.qua.measure] statement.
    These are the options which can be used inside the measure command as part of the ``counting`` process.

    -- Available with the OPD addon --
    """

    loc = ""

    @staticmethod
    def digital(
        target: QuaVariable[int],
        max_time: int,
        element_outputs: str = "",
    ) -> DigitalMeasureProcessCounting:
        """Performs counting from the attached OPD. See [Time tagging](../../Guides/features.md#time-tagging).

        -- Available with the OPD addon --

        Args:
            target (QUA int): A QUA int which will get the number of
                pulses detected
            max_time (int): The time in which pulses are detected
                (Must be larger than the pulse duration)
            element_outputs (str): the outputs of an element from which
                to get ADC results
        """
        return DigitalMeasureProcessCounting(element_outputs, target, max_time)


class demod(_RealAccumulationMethod):
    """
    A class for specifying the demodulation processes in the [measure][qm.qua.measure] statement.
    """

    @classmethod
    def _get_return_func(cls) -> Type[BasicIntegration]:
        return DemodIntegration

    @classmethod
    def full(
        cls, iw: str, target: Union[QuaVariable[float], QuaArrayCell[float]], element_output: str = ""
    ) -> BasicIntegration:
        """
        Perform an ordinary demodulation.
        See [Full demodulation](../../Guides/features.md#full-demodulation).

        Args:
            iw (str): integration weights
            target (QUA variable): variable to which demod result is
                saved
            element_output (str): The output of an element from which to get the ADC data.
                Required for elements with `MWOutput` or with multiple `outputs`.
                Optional otherwise.
        """
        return super().full(iw, target, element_output)

    @classmethod
    def sliced(
        cls,
        iw: str,
        target: QuaArrayVariable[float],
        samples_per_chunk: int,
        element_output: str = "",
    ) -> BasicIntegration:
        """
        Perform a demodulation in which the demodulation process is split into chunks,
        and the value of each chunk is saved in an array cell.
        See [Sliced demodulation](../../Guides/features.md#sliced-demodulation).

        Args:
            iw (str): integration weights
            target (QUA array): variable to which demod result is saved
            samples_per_chunk (int): The number of ADC samples to be
                used for each chunk is this number times 4.
            element_output (str): The output of an element from which to get the ADC data.
                Required for elements with `MWOutput` or with multiple `outputs`.
                Optional otherwise.
        """
        return super().sliced(iw, target, samples_per_chunk, element_output)

    @classmethod
    def accumulated(
        cls,
        iw: str,
        target: QuaArrayVariable[float],
        samples_per_chunk: int,
        element_output: str = "",
    ) -> BasicIntegration:
        """
        Perform a demodulation in which the demodulation process is split into chunks,
        and the accumulated result is saved in each array cell.
        See [Accumulated demodulation](../../Guides/features.md#accumulated-demodulation).

        Args:
            iw (str): integration weights
            target (QUA array): variable to which demod result is saved
            samples_per_chunk (int): The number of ADC samples to be
                used for each chunk is this number times 4.
            element_output (str): The output of an element from which to get the ADC data.
                Required for elements with `MWOutput` or with multiple `outputs`.
                Optional otherwise.
        """
        return super().accumulated(iw, target, samples_per_chunk, element_output)

    @classmethod
    def moving_window(
        cls,
        iw: str,
        target: QuaArrayVariable[float],
        samples_per_chunk: int,
        chunks_per_window: int,
        element_output: str = "",
    ) -> BasicIntegration:
        """
        Perform a demodulation in which the demodulation process is split into chunks,
        and several chunks are accumulated and saved to each array cell, creating a moving window effect.
        See [Moving window demodulation](../../Guides/features.md#moving-window-demodulation).

        Args:
            iw (str): integration weights
            target (QUA array): variable to which demod result is saved
            samples_per_chunk (int): The number of ADC samples to be
                used for each chunk is this number times 4.
            chunks_per_window (int): The number of chunks to use in the
                moving window
            element_output (str): The output of an element from which to get the ADC data.
                Required for elements with `MWOutput` or with multiple `outputs`.
                Optional otherwise.
        """
        return super().moving_window(iw, target, samples_per_chunk, chunks_per_window, element_output)


class dual_demod(_DualAccumulationMethod):
    """
    A class for specifying the dual demod processes in the [measure][qm.qua.measure] statement.
    Dual demod allows demodulating two inputs simultaneously and summing up the result.
    """

    @classmethod
    def _get_return_func(cls) -> Type[DualMeasureProcess]:
        return DualDemodIntegration

    @overload
    @classmethod
    def full(
        cls,
        iw1: str,
        iw2: str,
        target: Union[QuaVariable[float], QuaArrayCell[float]],
    ) -> DualMeasureProcess:
        """
        Perform an ordinary dual demodulation.
        See [Dual demodulation](../../Guides/demod.md#dual-demodulation).

        Args:
            iw1 (str): integration weights to be applied to the I quadrature of a `MWOutput` element
                or `out1` for an element with multiple `outputs`.
            iw2 (str): integration weights to be applied to the Q quadrature of a `MWOutput` element
                or `out2` for an element with multiple `outputs`.
            target (QUA variable): variable to which the demod result is saved
        """
        return super().full(iw1, iw2, target)

    @overload
    @classmethod
    def full(
        cls,
        iw1: str,
        element_output1: str,
        iw2: str,
        element_output2: str,
        target: Union[QuaVariable[float], QuaArrayCell[float]],
    ) -> DualMeasureProcess:
        """
        Perform an ordinary dual demodulation.
        See [Dual demodulation](../../Guides/demod.md#dual-demodulation).

        Args:
            iw1 (str): integration weights to be applied to element_output1
            element_output1 (str): the output of an element from which to get ADC results
            iw2 (str): integration weights to be applied to element_output2
            element_output2 (str): the output of an element from which to get ADC results
            target (QUA variable): variable to which the demod result is saved
        """
        return super().full(iw1, element_output1, iw2, element_output2, target)

    @classmethod
    def full(cls, *args: Any, **kwargs: Any) -> DualMeasureProcess:
        """
        Perform an ordinary dual demodulation.
        See [Dual demodulation](../../Guides/demod.md#dual-demodulation).

        Args:
            args (Any): Accepts either 3 or 5 positional arguments: `iw1`, `iw2`, `target` or `iw1`, `element_output1`, `iw2`, `element_output2`, `target`:

                * iw1 (str): integration weights to be applied to element_output1
                * element_output1 (str): the output of an element from which to get ADC results
                * iw2 (str): integration weights to be applied to element_output2
                * element_output2 (str): the output of an element from which to get ADC results
                * target (QUA variable): variable to which the demod result is saved

            kwargs (Any): Accepts keyword arguments corresponding to the parameters.
        """
        return super().full(*args, **kwargs)


class integration(_RealAccumulationMethod):
    """
    A class for specifying the integration processes in the [measure][qm.qua.measure] statement.
    """

    @classmethod
    def _get_return_func(cls) -> Type[BasicIntegration]:
        return BareIntegration

    @classmethod
    def full(
        cls, iw: str, target: Union[QuaVariable[float], QuaArrayCell[float]], element_output: str = ""
    ) -> BasicIntegration:
        """
        Perform an ordinary integration.
        See [Full integration](../../Guides/features.md#full).

        Args:
            iw (str): integration weights
            target (QUA variable): variable to which demod result is
                saved
            element_output (str): The output of an element from which to get the ADC data.
                Required for elements with `MWOutput` or with multiple `outputs`.
                Optional otherwise.
        """
        return super().full(iw, target, element_output)

    @classmethod
    def sliced(
        cls,
        iw: str,
        target: QuaArrayVariable[float],
        samples_per_chunk: int,
        element_output: str = "",
    ) -> BasicIntegration:
        """
        Perform an integration in which the integration process is split into chunks,
        and the value of each chunk is saved in an array cell.
        See [Sliced integration](../../Guides/features.md#sliced).

        Args:
            iw (str): integration weights
            target (QUA array): variable to which demod result is saved
            samples_per_chunk (int): The number of ADC samples to be
                used for each chunk is this number times 4.
            element_output (str): The output of an element from which to get the ADC data.
                Required for elements with `MWOutput` or with multiple `outputs`.
                Optional otherwise.
        """
        return super().sliced(iw, target, samples_per_chunk, element_output)

    @classmethod
    def accumulated(
        cls,
        iw: str,
        target: QuaArrayVariable[float],
        samples_per_chunk: int,
        element_output: str = "",
    ) -> BasicIntegration:
        """
        Perform an integration in which the integration process is split into chunks,
        and the accumulated result is saved in each array cell.
        See [Accumulated integration](../../Guides/features.md#accumulated).

        Args:
            iw (str): integration weights
            target (QUA array): variable to which demod result is saved
            samples_per_chunk (int): The number of ADC samples to be
                used for each chunk is this number times 4.
            element_output (str): The output of an element from which to get the ADC data.
                Required for elements with `MWOutput` or with multiple `outputs`.
                Optional otherwise.
        """
        return super().accumulated(iw, target, samples_per_chunk, element_output)

    @classmethod
    def moving_window(
        cls,
        iw: str,
        target: QuaArrayVariable[float],
        samples_per_chunk: int,
        chunks_per_window: int,
        element_output: str = "",
    ) -> BasicIntegration:
        """
        Perform an integration in which the integration process is split into chunks,
        and several chunks are accumulated and saved to each array cell, creating a moving window effect.
        See [Moving window integration](../../Guides/features.md#moving-window).

        Args:
            iw (str): integration weights
            target (QUA array): variable to which demod result is saved
            samples_per_chunk (int): The number of ADC samples to be
                used for each chunk is this number times 4.
            chunks_per_window (int): The number of chunks to use in the
                moving window
            element_output (str): The output of an element from which to get the ADC data.
                Required for elements with `MWOutput` or with multiple `outputs`.
                Optional otherwise.
        """
        return super().moving_window(iw, target, samples_per_chunk, chunks_per_window, element_output)



from qm.type_hinting.config_types import DictQuaConfig, FullQuaConfig
from qm.type_hinting.general import Value, Number, NumberT, PathLike, NumpySupportedValue, NumpySupportedNumber

__all__ = [
    "FullQuaConfig",
    "DictQuaConfig",
    "Number",
    "Value",
    "PathLike",
    "NumberT",
    "NumpySupportedValue",
    "NumpySupportedNumber",
]



from typing import Union, Optional, cast

from qm.type_hinting import Number
from qm.grpc.qm.pb import inc_qua_config_pb2
from qm.type_hinting.config_types import MixerConfigType
from qm.program._dict_to_pb_converter.base_converter import BaseDictToPbConverter


class MixerCorrectionConverter(BaseDictToPbConverter[MixerConfigType, inc_qua_config_pb2.QuaConfig.CorrectionEntry]):
    def convert(self, input_data: MixerConfigType) -> inc_qua_config_pb2.QuaConfig.CorrectionEntry:
        # Correction entries are stored in a list (refer to the function call). Unlike other values in the controller config,
        # lists do not support the 'upsert' operation. When a list is updated, it fully replaces the one set during init mode.
        # Therefore, the fields of inc_qua_config_pb2.QuaConfig.CorrectionEntry can not be optional, as 'upsert' is not supported, only full replacement.

        default_schema: MixerConfigType = {"intermediate_frequency": 0, "lo_frequency": 0}
        data_with_defaults = self._apply_defaults(input_data, default_schema=default_schema)

        # In "correction entry", all fields must be explicitly provided by the user in update mode. In init mode,
        # the correction field is mandatory, while the frequency parameters are assigned default values if not specified.
        # Therefore, after applying default values, all three fields should always be set.
        self._validate_required_fields(
            data_with_defaults, ["intermediate_frequency", "lo_frequency", "correction"], "mixer correction entry"
        )

        correction = inc_qua_config_pb2.QuaConfig.CorrectionEntry()

        correction.correction.CopyFrom(
            inc_qua_config_pb2.QuaConfig.Matrix(
                v00=data_with_defaults["correction"][0],
                v01=data_with_defaults["correction"][1],
                v10=data_with_defaults["correction"][2],
                v11=data_with_defaults["correction"][3],
            )
        )

        correction.frequencyNegative = data_with_defaults["intermediate_frequency"] < 0
        correction.frequency = abs(int(data_with_defaults["intermediate_frequency"]))
        if self._capabilities.supports_double_frequency:
            correction.frequencyDouble = abs(float(data_with_defaults["intermediate_frequency"]))

        correction.loFrequency = int(data_with_defaults["lo_frequency"])
        if self._capabilities.supports_double_frequency:
            correction.loFrequencyDouble = float(data_with_defaults["lo_frequency"])

        return correction

    def deconvert(self, output_data: inc_qua_config_pb2.QuaConfig.CorrectionEntry) -> MixerConfigType:
        frequency: Optional[Union[int, float]]
        lo_frequency: Optional[Union[int, float]]

        if output_data.frequencyDouble:
            frequency = output_data.frequencyDouble
        else:
            frequency = output_data.frequency

        if output_data.frequencyNegative is True:
            assert (
                frequency is not None
            )  # Mypy thinks it can be None, but it can't really (frequency has a default value)
            frequency = -frequency

        if output_data.loFrequencyDouble:
            lo_frequency = output_data.loFrequencyDouble
        else:
            lo_frequency = output_data.loFrequency

        correction_as_dict = cast(
            MixerConfigType,
            {
                "intermediate_frequency": frequency,
                "lo_frequency": lo_frequency,
                "correction": self._convert_matrix(output_data.correction),
            },
        )
        return correction_as_dict

    @staticmethod
    def _convert_matrix(matrix: inc_qua_config_pb2.QuaConfig.Matrix) -> tuple[Number, Number, Number, Number]:
        return matrix.v00, matrix.v01, matrix.v10, matrix.v11



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: qm/pb/compiler.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'qm/pb/compiler.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from qm.grpc.qm.pb import inc_qua_pb2 as qm_dot_pb_dot_inc__qua__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from qm.grpc.qm.pb import qm_manager_pb2 as qm_dot_pb_dot_qm__manager__pb2
from qm.grpc.qm.pb import general_messages_pb2 as qm_dot_pb_dot_general__messages__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14qm/pb/compiler.proto\x12\x10qm.grpc.compiler\x1a\x13qm/pb/inc_qua.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\x16qm/pb/qm_manager.proto\x1a\x1cqm/pb/general_messages.proto\"K\n\tQuaValues\x12\x11\n\tint_value\x18\x01 \x01(\r\x12\x14\n\x0c\x64ouble_value\x18\x02 \x01(\x01\x12\x15\n\rboolean_value\x18\x03 \x01(\x08\"I\n\x0e\x43ompileRequest\x12(\n\x07program\x18\x01 \x01(\x0b\x32\x17.qm.grpc.qua.QuaProgram\x12\r\n\x05jobId\x18\x02 \x01(\t\"\xdf\x01\n\x0f\x43ompileResponse\x12\n\n\x02ok\x18\x01 \x01(\x08\x12\x33\n\x08messages\x18\x02 \x03(\x0b\x32!.qm.grpc.compiler.CompilerMessage\x12\x0e\n\x06\x62inary\x18\x03 \x01(\x0c\x12\x10\n\x08metadata\x18\x04 \x01(\t\x12;\n\x05\x64\x65\x62ug\x18\x1e \x03(\x0b\x32,.qm.grpc.compiler.CompileResponse.DebugEntry\x1a,\n\nDebugEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"Y\n\x0f\x43ompilerMessage\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x35\n\x05level\x18\x02 \x01(\x0e\x32&.qm.grpc.general_messages.MessageLevel\"S\n\x12ValidationResponse\x12=\n\x08messages\x18\x01 \x03(\x0b\x32+.qm.grpc.qm_manager.ConfigValidationMessage\"G\n\rDynamicConfig\x12\x0f\n\x07version\x18\x01 \x01(\r\x12%\n\x04root\x18\x02 \x01(\x0b\x32\x17.google.protobuf.StructB)\n\x10qm.grpc.compilerB\rCompilerProtoP\x01\xa2\x02\x03HLWb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'qm.pb.compiler_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\020qm.grpc.compilerB\rCompilerProtoP\001\242\002\003HLW'
  _globals['_COMPILERESPONSE_DEBUGENTRY']._loaded_options = None
  _globals['_COMPILERESPONSE_DEBUGENTRY']._serialized_options = b'8\001'
  _globals['_QUAVALUES']._serialized_start=147
  _globals['_QUAVALUES']._serialized_end=222
  _globals['_COMPILEREQUEST']._serialized_start=224
  _globals['_COMPILEREQUEST']._serialized_end=297
  _globals['_COMPILERESPONSE']._serialized_start=300
  _globals['_COMPILERESPONSE']._serialized_end=523
  _globals['_COMPILERESPONSE_DEBUGENTRY']._serialized_start=479
  _globals['_COMPILERESPONSE_DEBUGENTRY']._serialized_end=523
  _globals['_COMPILERMESSAGE']._serialized_start=525
  _globals['_COMPILERMESSAGE']._serialized_end=614
  _globals['_VALIDATIONRESPONSE']._serialized_start=616
  _globals['_VALIDATIONRESPONSE']._serialized_end=699
  _globals['_DYNAMICCONFIG']._serialized_start=701
  _globals['_DYNAMICCONFIG']._serialized_end=772
# @@protoc_insertion_point(module_scope)



import warnings
from typing import Optional

from qm.grpc.qm.pb import inc_qua_config_pb2
from qm.type_hinting.config_types import FullQuaConfig
from qm.api.models.capabilities import ServerCapabilities, offline_capabilities

from ..utils import deprecation_message
from ._dict_to_pb_converter import DictToQuaConfigConverter


def convert_msg_to_config(
    config: inc_qua_config_pb2.QuaConfig,
    capabilities: Optional[ServerCapabilities] = None,
) -> FullQuaConfig:
    capabilities = capabilities if capabilities is not None else offline_capabilities
    warnings.warn(
        deprecation_message(
            method="convert_msg_to_config",
            deprecated_in="1.2.4",
            removed_in="1.3",
            details="This function is removed, if you need an alternative, please contact QM.",
        ),
        DeprecationWarning,
    )
    converter = DictToQuaConfigConverter(capabilities)
    return converter.deconvert(config)



import copy
import time
import logging
import dataclasses
from pprint import PrettyPrinter
from typing import Any, Dict, List, Tuple, Union, Callable, Optional, Sequence, NamedTuple, cast

import grpc
from google.protobuf.json_format import ParseDict, MessageToDict

from qm.grpc.octave.v1 import api_pb2
from qm.utils.protobuf_utils import which_one_of
from qm.octave_sdk.batch import Batched, BatchSingleton
from qm.grpc.octave.v1.api_pb2_grpc import OctaveServiceStub
from qm.octave_sdk.connectivity.connectivity_util import (
    slot_index_to_panel_mapping,
    octave_module_to_module_name_mapping,
)

logger = logging.getLogger("qm")

"""
# OctaveClient
#
#   This is a gRPC based Octave. The standard setup is the whole Octave product.
#   Namely, we access the SOM on its motherboard via ethernet and from there
#   communicate with all the boards via the main PIC.
#
#   A temporary alternative is using a mini SOM eval board which is connected
#   to a mini motherboard, directly into the serial lines of its PIC.
#
#   The specific variation is passed to the constructor.
"""

module_type_to_class = {
    api_pb2.OctaveModule.OCTAVE_MODULE_RF_UPCONVERTER: "rf_up_conv",
    api_pb2.OctaveModule.OCTAVE_MODULE_RF_DOWNCONVERTER: "rf_down_conv",
    api_pb2.OctaveModule.OCTAVE_MODULE_SYNTHESIZER: "synth",
    api_pb2.OctaveModule.OCTAVE_MODULE_IF_DOWNCONVERTER: "if_down_conv",
    api_pb2.OctaveModule.OCTAVE_MODULE_MOTHERBOARD: "motherboard",
}


class DebugSetException(Exception):
    pass


def _build_grpc_channel(
    host: str,
    port: int,
    octave_name: Optional[str] = None,
    credentials: Optional[grpc.ChannelCredentials] = None,
    options: Optional[List[tuple[str, int]]] = None,
) -> grpc.Channel:
    target = f"{host}:{port}"

    # Default options for message size
    default_options = [
        ("grpc.max_receive_message_length", 100 * 1024 * 1024),
        ("grpc.max_send_message_length", 100 * 1024 * 1024),
    ]

    if options:
        default_options.extend(options)

    if credentials:
        channel = grpc.secure_channel(target, credentials, options=default_options)
    else:
        channel = grpc.insecure_channel(target, options=default_options)

    return channel


@dataclasses.dataclass
class MonitorData:
    temp: float
    errors: List[api_pb2.MonitorResponse.ModuleStatusError]


class MonitorResult:
    def __init__(self, results_pb: api_pb2.MonitorResponse):
        rf_upconverters: List[Optional[MonitorData]] = [None for _ in range(5)]
        rf_downconverters: List[Optional[MonitorData]] = [None for _ in range(2)]
        if_downconverters: List[Optional[MonitorData]] = [None for _ in range(2)]
        synthesizers: List[Optional[MonitorData]] = [None for _ in range(6)]
        motherboard: List[Optional[MonitorData]] = [None for _ in range(1)]
        self.modules: Dict[api_pb2.OctaveModule, List[Optional[MonitorData]]] = {
            api_pb2.OctaveModule.OCTAVE_MODULE_RF_UPCONVERTER: rf_upconverters,
            api_pb2.OctaveModule.OCTAVE_MODULE_RF_DOWNCONVERTER: rf_downconverters,
            api_pb2.OctaveModule.OCTAVE_MODULE_IF_DOWNCONVERTER: if_downconverters,
            api_pb2.OctaveModule.OCTAVE_MODULE_SYNTHESIZER: synthesizers,
            api_pb2.OctaveModule.OCTAVE_MODULE_SOM: motherboard,
        }
        for module in results_pb.modules:
            self.modules[module.module.type][module.module.index - 1] = MonitorData(
                temp=module.temperature, errors=list(module.errors)
            )

    def __repr__(self) -> str:
        pp = PrettyPrinter(width=150)
        return pp.pformat(self.modules)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, MonitorResult):
            return False
        # Basically compare all elements except temp
        self_copy = copy.deepcopy(self)
        self._clear_temp_values(self_copy)
        other_copy = copy.deepcopy(other)
        self._clear_temp_values(other_copy)

        return self_copy.modules == other_copy.modules

    def _clear_temp_values(self, monitor_result: "MonitorResult") -> None:
        for module_list in monitor_result.modules.values():
            for monitor_data in module_list:
                if monitor_data is not None:
                    monitor_data.temp = 0

    @property
    def temperatures(self) -> Dict[str, float]:
        res = {}
        for module_type, module_list in self.modules.items():
            for index, monitor_data in enumerate(module_list):
                if monitor_data is not None:
                    key = f"{octave_module_to_module_name_mapping[module_type]} {slot_index_to_panel_mapping(index + 1, module_type)}"
                    res[key] = monitor_data.temp
        return res


class ExploreResult:
    def __init__(self, results_pb: api_pb2.ExploreResponse):
        rf_upconverters: List[Optional[str]] = [None for _ in range(5)]
        rf_downconverters: List[Optional[str]] = [None for _ in range(2)]
        if_downconverters: List[Optional[str]] = [None for _ in range(2)]
        synthesizers: List[Optional[str]] = [None for _ in range(6)]
        motherboard: List[Optional[str]] = [None for _ in range(1)]
        self.modules = {
            api_pb2.OctaveModule.OCTAVE_MODULE_RF_UPCONVERTER: rf_upconverters,
            api_pb2.OctaveModule.OCTAVE_MODULE_RF_DOWNCONVERTER: rf_downconverters,
            api_pb2.OctaveModule.OCTAVE_MODULE_IF_DOWNCONVERTER: if_downconverters,
            api_pb2.OctaveModule.OCTAVE_MODULE_SYNTHESIZER: synthesizers,
            api_pb2.OctaveModule.OCTAVE_MODULE_MOTHERBOARD: motherboard,
        }
        for module in results_pb.modules:
            self.modules[module.module.type][module.module.index - 1] = module.id

    def __repr__(self) -> str:
        module_types = [
            api_pb2.OctaveModule.OCTAVE_MODULE_RF_UPCONVERTER,
            api_pb2.OctaveModule.OCTAVE_MODULE_RF_DOWNCONVERTER,
            api_pb2.OctaveModule.OCTAVE_MODULE_IF_DOWNCONVERTER,
            api_pb2.OctaveModule.OCTAVE_MODULE_SYNTHESIZER,
        ]

        m_id = self.modules[api_pb2.OctaveModule.OCTAVE_MODULE_MOTHERBOARD][0]
        if m_id is None or m_id == "":
            res = "MOTHERBOARD [\x1b[38;5;208m?? ?\x1b[0m]\n"
        else:
            res = f"MOTHERBOARD [\x1b[38;5;154m{m_id}\x1b[0m]\n"

        res += "    RF_UPCONVs          RF_DOWNCONVs        IF_DOWNCONVs        " "SYNTHESIZERs\n"
        for index in range(6):
            res += "     "
            for t in module_types:
                if index < len(self.modules[t]):
                    m_id = self.modules[t][index]
                    if m_id is None:
                        res += f"\x1b[38;5;241m{index + 1}. {'---':17s}\x1b[0m"
                    elif m_id == "":
                        res += f"{index + 1}. \x1b[38;5;208m{'???':17s}\x1b[0m"
                    else:
                        res += f"{index + 1}. \x1b[38;5;154m{m_id:17s}\x1b[0m"
                else:
                    res += f"{'':20s}"
            res += "\n"
        return res


MetadataValue = Union[str, bytes]


class _ClientCallDetails(NamedTuple):
    method: str
    timeout: Optional[float]
    metadata: Optional[Sequence[Tuple[str, MetadataValue]]]
    credentials: Optional[grpc.CallCredentials]
    wait_for_ready: Optional[bool]


class OctaveMetadataInterceptor(grpc.UnaryUnaryClientInterceptor):
    """Interceptor to add custom metadata to all gRPC calls"""

    def __init__(self, metadata: Dict[str, str]) -> None:
        self._metadata: List[Tuple[str, str]] = list(metadata.items())

    def intercept_unary_unary(
        self,
        continuation: Callable[..., Any],
        client_call_details: grpc.ClientCallDetails,
        request: Any,
    ) -> Any:
        # Add custom metadata to the call
        new_metadata: List[Tuple[str, Union[str, bytes]]] = list(client_call_details.metadata or [])
        new_metadata.extend(self._metadata)

        new_details = _ClientCallDetails(
            method=client_call_details.method,
            timeout=client_call_details.timeout,
            metadata=new_metadata,
            credentials=client_call_details.credentials,
            wait_for_ready=client_call_details.wait_for_ready,
        )

        return continuation(new_details, request)


class OctaveClient(Batched):
    def __init__(
        self,
        host: str,
        port: int,
        octave_name: Optional[str] = None,
        connection_headers: Optional[Dict[str, str]] = None,
        credentials: Optional[grpc.ChannelCredentials] = None,
        options: Optional[List[tuple[str, int]]] = None,
    ) -> None:
        self._host = host
        self._port = port
        self._octave_name = octave_name or "null"
        self._headers = connection_headers or {}
        super().__init__()

        # Create the channel
        self._channel = _build_grpc_channel(host, port, octave_name, credentials, options)

        # Create interceptor for headers if needed
        if self._headers:
            interceptor = OctaveMetadataInterceptor(self._headers)
            self._channel = grpc.intercept_channel(self._channel, interceptor)

        # Add octave service header if specified
        if octave_name is not None:
            octave_interceptor = OctaveMetadataInterceptor({"x-grpc-service": octave_name})
            self._channel = grpc.intercept_channel(self._channel, octave_interceptor)

        self._service = OctaveServiceStub(self._channel)  # type: ignore[no-untyped-call]

    def __hash__(self) -> int:
        return hash((self._octave_name, self._host, self._port) + tuple(sorted(self._headers.items())))

    @property
    def name(self) -> str:
        return self._octave_name

    def __del__(self) -> None:
        self._service = None  # type: ignore[assignment]
        self._channel.close()

    def _control(self, w_data: bytes = b"", r_length: int = 0) -> api_pb2.ControlResponse:
        control_request = api_pb2.ControlRequest(w_data=w_data, r_length=r_length)
        return cast(api_pb2.ControlResponse, self._service.Control(control_request))

    def _format_cached_modules(self, modules: api_pb2.AquireResponse) -> None:
        cached_modules = {}
        for update in modules.state.updates:
            module_type, message = which_one_of(update, "update")
            if isinstance(
                message,
                (api_pb2.RFUpConvUpdate, api_pb2.RFDownConvUpdate, api_pb2.IFDownConvUpdate, api_pb2.SynthUpdate),
            ):
                update_id = message.index
            else:
                update_id = 0

            cache_key = (update_id, module_type)
            cached_modules[cache_key] = update
        BatchSingleton().set_cached_modules(self, cached_modules)

    def _start_batch_callback(self) -> None:
        self._format_cached_modules(self.acquire_all_modules())

    def _end_batch_callback(self) -> None:
        self._send_update(list(BatchSingleton().get_cached_updates(self).values()))

    def update(self, updates: List[api_pb2.SingleUpdate]) -> None:
        if BatchSingleton().is_batch_mode:
            for update in updates:
                self._cache_update(update)
        else:
            self._send_update(updates)

    def _cache_update(self, update: api_pb2.SingleUpdate) -> None:
        module_type, message = which_one_of(update, "update")

        if isinstance(
            message, (api_pb2.RFUpConvUpdate, api_pb2.RFDownConvUpdate, api_pb2.IFDownConvUpdate, api_pb2.SynthUpdate)
        ):
            update_id = message.index
        else:
            update_id = 0

        cache_key = (update_id, module_type)
        current_updates = BatchSingleton().get_cached_updates(self)
        if current_updates.get(cache_key):
            previous_cached_update = MessageToDict(getattr(current_updates.get(cache_key), module_type))
        else:
            previous_cached_update = {}

        assert message is not None
        previous_cached_update.update(MessageToDict(message))

        new_update = api_pb2.SingleUpdate()
        ParseDict(previous_cached_update, getattr(new_update, module_type))
        current_updates[cache_key] = new_update
        BatchSingleton().set_cached_updates(self, current_updates)

    def _send_update(self, updates: List[api_pb2.SingleUpdate]) -> api_pb2.UpdateResponse:
        update_request = api_pb2.UpdateRequest(updates=updates)
        response = self._service.Update(update_request)
        if not response.success:
            raise Exception(f"Octave update failed: {response.error_message}")
        return cast(api_pb2.UpdateResponse, response)

    def debug_request_clock_print(self) -> None:
        self._control(
            w_data=bytes([0xFF, 0xB9]),
            r_length=1,
        )

    def debug_set(
        self,
        monitor_enabled: Optional[bool] = None,
        monitor_timeout: Optional[int] = None,
        monitor_print_rate: Optional[int] = None,
        monitor_update_fan: Optional[bool] = None,
        uart_debug_mode: Optional[bool] = None,
        print_updates: Optional[bool] = None,
        min_fan_speed: Optional[bool] = None,
        min_temp: Optional[int] = None,
        max_temp_modules: Optional[int] = None,
        max_temp_fpga: Optional[int] = None,
    ) -> None:
        if monitor_timeout is not None:
            if monitor_timeout < 1 or monitor_timeout > 15:
                print("OctaveClientBase.debug_set   ERROR    monitor_timeout should be" " 1..15")
                return
        else:
            monitor_timeout = 0x00

        if monitor_print_rate is not None:
            if monitor_print_rate < 0 or monitor_print_rate > 255:
                print(
                    "OctaveClientBase.debug_set   ERROR    monitor_print_rate should either 0 (no printings) or 1..255"
                )
                return

        activate = 0x00
        state = 0x00

        if monitor_enabled is not None:
            activate |= 0x01
            state |= 0x01 if monitor_enabled else 0x00

        if uart_debug_mode is not None:
            activate |= 0x02
            state |= 0x02 if uart_debug_mode else 0x00

        if print_updates is not None:
            activate |= 0x04
            state |= 0x04 if print_updates else 0x00

        if monitor_print_rate is not None:
            activate |= 0x08
        else:
            monitor_print_rate = 0

        if monitor_update_fan is not None:
            activate |= 0x10
            state |= 0x10 if monitor_update_fan else 0x00

        if min_fan_speed is not None:
            activate |= 0x20
            min_fan_speed = min(int(min_fan_speed), 31)  # type: ignore[assignment]
            # Not sure what's going on here, where min_fan_speed is actually a bool
            state |= (min_fan_speed & 1) << 5
            monitor_timeout |= (min_fan_speed & 0x1E) << 3

        if min_temp is not None:
            if min_temp < 1 or min_temp > 60:
                print("OctaveClientBase.debug_set   ERROR    min_temp should be between 1 to 60")
                return
            activate |= 0x40
        else:
            min_temp = 0x00

        if (max_temp_modules and not max_temp_fpga) or (max_temp_fpga and not max_temp_modules):
            logger.error("max_temp_modules and max_temp_fpga must come in paris")
            return

        if max_temp_modules and max_temp_fpga:
            if max_temp_modules < 1 or max_temp_modules > 65:
                print("OctaveClientBase.debug_set   ERROR    max_temp_modlues should be between 1 to 65")
                return
            if max_temp_fpga < 1 or max_temp_fpga > 75:
                print("OctaveClientBase.debug_set   ERROR    max_temp_fpga should be between 1 to 75")
                return
            activate |= 0x80
        else:
            max_temp_modules = 0
            max_temp_fpga = 0

        control_payload_old: bytes = bytes(
            [
                0xFF,
                0xFF,
                activate,
                state,
                monitor_timeout,
                monitor_print_rate,
            ]
        )

        control_payload_new = control_payload_old + bytes(
            [
                min_temp,
                max_temp_modules,
                max_temp_fpga,
            ]
        )

        # Try new format
        res = (
            self._control(
                w_data=control_payload_new,
                r_length=1,
            )
        ).r_data
        res_int = int.from_bytes(res, "little")
        if res_int == api_pb2.ControlResponse.RdataDebug.RDATA_DEBUG_SUCCESS_RESPONSE:
            return

        # Fall back to old format
        res = (
            self._control(
                w_data=control_payload_old,
                r_length=1,
            )
        ).r_data
        res_int = int.from_bytes(res, "little")
        if res_int != api_pb2.ControlResponse.RdataDebug.RDATA_DEBUG_SUCCESS_RESPONSE:
            raise DebugSetException("Failed to set debug params")

    def save_modules(
        self,
        m_id: Optional[str] = None,
        module_refs: Optional[List[api_pb2.ModuleReference]] = None,
        overwrite: bool = True,
    ) -> api_pb2.SaveResponse:
        module_id = m_id or "default"
        module_references = module_refs if module_refs else []
        save_request = api_pb2.SaveRequest(
            id=module_id, modules=module_references, overwrite=overwrite, timestamp=int(time.time())
        )

        return cast(api_pb2.SaveResponse, self._service.Save(save_request))

    def _fetch_acquire_from_cache(self, modules: List[api_pb2.ModuleReference]) -> api_pb2.AquireResponse:
        res = api_pb2.AquireResponse(state=api_pb2.UpdateRequest(updates=[]))
        for module in modules:
            module_type = module_type_to_class[module.type]
            if module_type != "motherboard":
                update_id = module.index
            else:
                update_id = 0

            cache_key = (update_id, module_type)
            cached_modules = BatchSingleton().get_cached_modules(self)
            cached_updates = BatchSingleton().get_cached_updates(self)

            if cache_key in cached_modules:
                cached_module = MessageToDict(cached_modules[cache_key])
                cached_update = MessageToDict(cached_updates.get(cache_key, api_pb2.SingleUpdate()))

                cached_module.update(cached_update)
                new_update = api_pb2.SingleUpdate()
                ParseDict(cached_module, new_update)
                res.state.updates.append(new_update)
        return res

    def aquire_modules(
        self, modules: List[api_pb2.ModuleReference], use_cache: bool = True
    ) -> Union[api_pb2.AquireResponse, api_pb2.SingleUpdate]:
        res = self.acquire_modules(modules, use_cache)

        if len(modules) == 1 and len(res.state.updates) == 1:
            return res.state.updates[0]

        return res

    def acquire_modules(
        self, modules: Sequence[api_pb2.ModuleReference] = tuple(), use_cache: bool = True
    ) -> api_pb2.AquireResponse:
        modules = list(modules)
        if BatchSingleton().is_batch_mode:
            return self._fetch_acquire_from_cache(modules)
        else:
            request = api_pb2.AquireRequest(modules=modules, use_cache=use_cache)
            res = self._service.Aquire(request)
        return cast(api_pb2.AquireResponse, res)

    def acquire_all_modules(self, use_cache: bool = True) -> api_pb2.AquireResponse:
        return self.acquire_modules(use_cache=use_cache)

    def acquire_module(self, module: api_pb2.ModuleReference, use_cache: bool = True) -> api_pb2.SingleUpdate:
        return self.acquire_modules([module], use_cache=use_cache).state.updates[0]

    def recall(self, m_id: Optional[str] = None) -> api_pb2.RecallResponse:
        recall_request = api_pb2.RecallRequest(id="default" if m_id is None else m_id)
        return cast(api_pb2.RecallResponse, self._service.Recall(recall_request))

    def configs(self) -> List[api_pb2.SaveInfo]:
        return cast(List[api_pb2.SaveInfo], self._service.List(api_pb2.ListRequest()).save_infos)

    def version(self) -> api_pb2.GetVersionResponse:
        return cast(api_pb2.GetVersionResponse, self._service.GetVersion(api_pb2.GetVersionRequest()))

    def monitor(self, sense_only: bool = True) -> MonitorResult:
        monitor_request = api_pb2.MonitorRequest(sense_only=sense_only)
        return MonitorResult(self._service.Monitor(monitor_request))

    def explore(self) -> ExploreResult:
        return ExploreResult(self._service.Explore(api_pb2.ExploreRequest()))

    def identify(self) -> api_pb2.IdentifyResponse:
        return cast(api_pb2.IdentifyResponse, self._service.Identify(api_pb2.IdentifyRequest()))

    def reset(self) -> api_pb2.ResetResponse:
        return cast(api_pb2.ResetResponse, self._service.Reset(api_pb2.ResetRequest()))

    @property
    def host(self) -> str:
        return self._host

    @property
    def port(self) -> int:
        return self._port

    @property
    def octave_name(self) -> str:
        return self._octave_name



import warnings

from qm.utils import deprecation_message


def create_capabilities_container(qua_implementation: None) -> None:
    """This is here just to check if the CI passes, SW-validation imports this function and call it with None"""
    # TODO - tell validation to remove this call
    warnings.warn(
        deprecation_message(
            "create_capabilities_container",
            "1.2.4",
            "1.2.5",
            "This function does nothing, and will be removed in the next version. Please remove it",
        )
    )
    return None



from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class GetInfoRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetInfoResponse(_message.Message):
    __slots__ = ("implementation", "capabilities")
    IMPLEMENTATION_FIELD_NUMBER: _ClassVar[int]
    CAPABILITIES_FIELD_NUMBER: _ClassVar[int]
    implementation: ImplementationDetails
    capabilities: _containers.RepeatedScalarFieldContainer[str]
    def __init__(self, implementation: _Optional[_Union[ImplementationDetails, _Mapping]] = ..., capabilities: _Optional[_Iterable[str]] = ...) -> None: ...

class ImplementationDetails(_message.Message):
    __slots__ = ("name", "version", "url", "proto_version")
    NAME_FIELD_NUMBER: _ClassVar[int]
    VERSION_FIELD_NUMBER: _ClassVar[int]
    URL_FIELD_NUMBER: _ClassVar[int]
    PROTO_VERSION_FIELD_NUMBER: _ClassVar[int]
    name: str
    version: str
    url: str
    proto_version: str
    def __init__(self, name: _Optional[str] = ..., version: _Optional[str] = ..., url: _Optional[str] = ..., proto_version: _Optional[str] = ...) -> None: ...



import ssl
import json
import logging
import warnings
from pathlib import Path
from dataclasses import field, dataclass
from typing import Any, Dict, List, Union, Mapping, Iterable, Optional, TypedDict, Collection

import marshmallow
import packaging.version

import qm
from qm.api.v2.qm_api import QmApi
from qm.user_config import UserConfig
from qm.utils import deprecation_message
from qm.api.frontend_api import FrontendApi
from qm.program import Program, load_config
from qm.type_hinting.general import PathLike
from qm.api.v2.job_api.job_api import JobData
from qm.quantum_machine import QuantumMachine
from qm.api.models.debug_data import DebugData
from qm.jobs.simulated_job import SimulatedJob
from qm.logging_utils import set_logging_level
from qm.octave_sdk.octave import OctaveDetails
from qm.api.v2.job_api import JobApi, JobStatus
from qm.api.server_detector import detect_server
from qm.simulate.interface import SimulationConfig
from qm.version import __version__ as qm_qua_version
from qm.api.job_result_api import JobResultServiceApi
from qm.persistence import BaseStore, SimpleFileStore
from qm.api.models.server_details import ServerDetails
from qm.grpc.qm.pb import frontend_pb2, inc_qua_config_pb2
from qm.utils.config_utils import get_controller_pb_config
from qm.octave import QmOctaveConfig, AbstractCalibrationDB
from qm.api.v2.job_api.simulated_job_api import SimulatedJobApi
from qm.utils.general_utils import is_debug, parse_proto_version
from qm._octaves_container import load_config_from_calibration_db
from qm.program._qua_config_schema import validate_config_capabilities
from qm.api.simulation_api import SimulationApi, create_simulation_request
from qm.type_hinting.config_types import FullQuaConfig, ControllerQuaConfig
from qm.octave.octave_manager import OctaveManager, prep_config_for_calibration
from qm.api.v2.qmm_api import Controller, ControllerBase, QmmApiWithDeprecations
from qm.exceptions import QmmException, ConfigSchemaError, ConfigValidationException
from qm.api.models.compiler import CompilerOptionArguments, standardize_compiler_params
from qm.api.models.capabilities import QopCaps, Capability, ServerCapabilities, offline_capabilities

from ._stream_results import StreamsManager
from .program._dict_to_pb_converter import DictToQuaConfigConverter

logger = logging.getLogger(__name__)

Version = TypedDict(
    "Version",
    {"qm-qua": str, "QOP": str, "OPX": str, "client": str, "server": str},
    total=False,
)


@dataclass
class DevicesVersion:
    gateway: str
    controllers: Dict[str, str]
    qm_qua: str
    octaves: Dict[str, str]
    QOP: Optional[str] = field(init=False)

    def __post_init__(self) -> None:
        self.QOP = _get_qop_version(self.gateway)


SERVER_TO_QOP_VERSION_MAP = {
    "2.40-144e7bb": "2.0.0",
    "2.40-82d4afc": "2.0.1",
    "2.40-8521884": "2.0.2",
    "2.40-3a0d7f1": "2.0.3",
    "2.50-1a24163": "2.1.3",
    "2.60-5ba458f": "2.2.0",
    "2.60-b62e6b6": "2.2.1",
    "2.60-0b17cac": "2.2.2",
    "2.70-7abf0e0": "2.4.0",
    "2.70-ed75211": "2.4.2",
    "2.70-0b3bd6b": "2.4.4",
    "2.70-cb7c3ee": "2.5.0",
    "2.70-a0974ba": "2.6.0",
    "a6f8bc5": "3.1.0",
    "fcdfb69": "3.1.1",
    "3.0-beta-78b5e00": "3.2.0",
    "3.0-beta-b43e229": "3.2.2",
    "3.0-beta-ba2d179": "3.2.3",
    "3.0-beta-eddf92b": "3.2.4",
    "3.0-beta-a3d43d5": "3.3.0",
    "3.0-beta-a59f2ea": "3.3.1",
    "3.0-beta-f47a556": "3.4.0",
    "3.0-beta-c2dd405": "3.4.1",
    "3.0-beta-4e4aa38": "3.5.0",
    "3.0-beta-10692.d6b91d0": "3.5.1",
    "3.6.0-11345.aa8a9fd": "3.6.0",
    "3.6.0-12658.702d99c": "3.6.1",
    "3.6.0-13160.a7d6ed8": "3.6.2",
}


def _get_qop_version(server_version: str) -> Optional[str]:
    """Convert a QOP server version string to a semantic version.

    For legacy server versions, uses a hardcoded lookup map. For newer versions,
    parses the server version string by replacing hyphens with local version
    separators and extracting the base version.

    Args:
        server_version: The raw version string reported by the QOP server
            (e.g. ``"2.40-144e7bb"`` or ``"3.7.0-RC1"``).

    Returns:
        The base semantic version string (e.g. ``"2.0.0"``, ``"3.7.0"``),
        or ``None`` if the version string cannot be parsed.
    """
    if server_version in SERVER_TO_QOP_VERSION_MAP:
        return SERVER_TO_QOP_VERSION_MAP[server_version]
    try:
        return packaging.version.parse(server_version.replace("-", "+")).base_version
    except packaging.version.InvalidVersion:
        return None


@dataclass
class Devices:
    controllers: Mapping[str, ControllerBase]
    octaves: Dict[str, OctaveDetails]


class QuantumMachinesManager:
    def __init__(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        *,
        cluster_name: Optional[str] = None,
        timeout: Optional[float] = None,
        log_level: Union[int, str] = logging.INFO,
        connection_headers: Optional[Dict[str, str]] = None,
        add_debug_data: bool = False,
        credentials: Optional[ssl.SSLContext] = None,
        store: Optional[BaseStore] = None,
        file_store_root: Optional[str] = None,
        octave: Optional[QmOctaveConfig] = None,
        octave_calibration_db_path: Optional[Union[PathLike, AbstractCalibrationDB]] = None,
        follow_gateway_redirections: bool = True,
        async_follow_redirects: bool = False,
        async_trust_env: bool = True,
    ):
        """
        Args:
            host (string): Host where to find the QM orchestrator. If ``None``, local settings are used.
            port: Port where to find the QM orchestrator. If None, local settings are used.
            cluster_name (string): The name of the cluster. Requires redirection between devices.
            timeout (float): The timeout, in seconds, for detecting the qmm and most other gateway API calls. Default is 60.
            log_level (Union[int, string]): The logging level for the connection instance. Defaults to `INFO`. Please check `logging` for available options.
            octave (QmOctaveConfig): The configuration for the Octave devices. Deprecated from QOP 2.4.0.
            octave_calibration_db_path (PathLike): The path for storing the Octave's calibration database. It can also be a calibration database which is an instance of `AbstractCalibrationDB`.
            follow_gateway_redirections (bool): If True (default), the client will follow redirections to find a QuantumMachinesManager and Octaves. Otherwise, it will only connect to the given host and port.
            async_follow_redirects (bool): If False (default), async httpx will not follow redirections, relevant only in case follow_gateway_redirections is True.
            async_trust_env (bool): If True (default), async httpx will read the environment variables for settings as proxy settings, relevant only in case follow_gateway_redirections is True.
        """
        set_logging_level(log_level)
        self._user_config = UserConfig.create_from_file()
        self._port = port
        host = host or self._user_config.manager_host or ""
        octave_calibration_db_path = octave_calibration_db_path or self._user_config.octave_calibration_db_path
        if host is None:
            message = "Failed to connect to QuantumMachines server. No host given."
            logger.error(message)
            raise QmmException(message)

        if file_store_root is not None:
            warnings.warn(
                deprecation_message(
                    "file_store_root",
                    "1.2.3",
                    "2.0.0",
                    "This parameter is going to be removed, remove it from you initialization",
                ),
                DeprecationWarning,
                stacklevel=2,
            )
        else:
            file_store_root = "."

        if store is not None:
            warnings.warn(
                deprecation_message(
                    "store",
                    "1.2.3",
                    "2.0.0",
                    "This parameter is going to be removed, remove it from you initialization",
                ),
                DeprecationWarning,
                stacklevel=2,
            )

        self._cluster_name = cluster_name
        self._store = store if store else file_store_root
        self._server_details = self._initialize_connection(
            host=host,
            port=port,
            timeout=timeout,
            add_debug_data=add_debug_data,
            connection_headers=connection_headers,
            credentials=credentials,
            follow_gateway_redirections=follow_gateway_redirections,
            async_follow_redirects=async_follow_redirects,
            async_trust_env=async_trust_env,
        )
        self._raise_warning_for_outdated_qm_qua_version()

        if self._server_details.octaves:
            if octave is None:
                octave = QmOctaveConfig()
            if octave.get_devices():
                warnings.warn(
                    "QMM was opened with OctaveConfig. Please note that from QOP2.4.0 the octave devices "
                    "are managed by the cluster setting in the QM-app. It is recommended to remove the "
                    "OctaveConfig from the QMM instantiation.",
                    category=DeprecationWarning,
                )
            else:
                for name, info in self._server_details.octaves.items():
                    octave.add_device(name, info)
        elif octave is not None:
            opx_headers = self._server_details.connection_details.headers
            for _, device in octave.get_devices().items():
                device.headers = {**opx_headers, **device.headers}

        self._caps = self._server_details.capabilities
        self._frontend = FrontendApi(self._server_details.connection_details)
        self._simulation_api = SimulationApi(self._server_details.connection_details)
        self._octave_config = octave
        self._octave_manager_cached: Optional[OctaveManager] = None

        self._perform_octaves_healthcheck_if_needed()

        self._api = None
        if self._caps.supports(QopCaps.qop3):
            self._api = QmmApiWithDeprecations(
                self._server_details.connection_details,
                capabilities=self._caps,
                octave_config=self._octave_config,
                octave_manager=self._octave_manager,
            )
            self._api.perform_healthcheck()
        else:
            strict = self._user_config.strict_healthcheck is not False
            self._frontend.healthcheck(strict)

        if octave_calibration_db_path is not None and self._octave_config:
            if self._octave_config.calibration_db is not None:
                raise QmmException(
                    "Duplicate calibration_db path detected, please set the calibration db only through the QMM."
                )
            else:
                self._octave_config.set_calibration_db_without_warning(octave_calibration_db_path)

    def _initialize_connection(
        self,
        host: str,
        port: Optional[int],
        timeout: Optional[float],
        add_debug_data: bool,
        credentials: Optional[ssl.SSLContext],
        connection_headers: Optional[Dict[str, str]],
        follow_gateway_redirections: bool,
        async_follow_redirects: bool,
        async_trust_env: bool,
    ) -> ServerDetails:
        server_details = detect_server(
            cluster_name=self._cluster_name,
            user_token=self._user_config.user_token,
            ssl_context=credentials,
            host=host,
            port_from_user_config=self._user_config.manager_port,
            user_provided_port=port,
            add_debug_data=add_debug_data,
            timeout=timeout,
            extra_headers=connection_headers,
            follow_gateway_redirections=follow_gateway_redirections,
            async_follow_redirects=async_follow_redirects,
            async_trust_env=async_trust_env,
        )
        offline_capabilities.set_from_implementation(server_details.qua_implementation)
        return server_details

    @property
    def store(self) -> BaseStore:
        warnings.warn(
            deprecation_message("qmm.store", "1.2.3", "2.0.0"),
            DeprecationWarning,
            stacklevel=2,
        )
        if isinstance(self._store, str):
            return SimpleFileStore(self._store)
        return self._store

    @property
    def capabilities(self) -> ServerCapabilities:
        return self._caps

    @property
    def octave_manager(self) -> OctaveManager:
        # warnings.warn(
        #     "Do not use OctaveManager, it will be removed in the next version", DeprecationWarning, stacklevel=2
        # )
        return self._octave_manager

    @property
    def _octave_manager(self) -> OctaveManager:
        """
        There are two flows of initialization:
        1. (Healthcheck is supported in GW) - in this case, we first initialize the qmm and then, when we see that the
           Octaves are healthy, we initialize the Octave manager (when needed).
        2. (Healthcheck is not supported in GW) - in this case, we first initialize the qmm and create the
           Octave manager, so we will be able to use it in the healthcheck.
        """
        if self._octave_manager_cached is None:
            self._octave_manager_cached = OctaveManager(self._octave_config, self, self._caps)
        return self._octave_manager_cached

    @property
    def cluster_name(self) -> str:
        return self._cluster_name or "any"

    def perform_healthcheck(self, strict: bool = True) -> None:
        """Perform a health check against the QM server.

        Args:
            strict: Will raise an exception if health check failed
        """
        if self._api:
            self._api.perform_healthcheck()
        else:
            self._frontend.healthcheck(strict)
        self._perform_octaves_healthcheck_if_needed()

    def _perform_octaves_healthcheck_if_needed(self) -> None:
        if self._octave_config is not None and not self._caps.supports(QopCaps.octave_management):
            for octave in self._octave_config.get_devices():
                octave_client = self._octave_manager.get_client(octave)
                octave_client.perform_healthcheck()

    @staticmethod
    def _get_local_proto_version() -> str:
        version_path = Path(qm.__file__).parent / "grpc" / "VERSION"
        return version_path.read_text().strip()

    def _raise_warning_for_outdated_qm_qua_version(self) -> None:
        """
        Raises a warning if the current qm-qua version is outdated compared to the proto used by the server.
        For example, the qm-qua used is 1.2.2, while the qop version is a newer one, which uses a proto compatible with qm-qua 1.2.3.
        We do this check by comparing between the local proto version and the proto version returned by the gateway (which contains the qm-qua version it is compatible with).
        """
        if self._server_details.proto_version is not None:
            local_proto_version = parse_proto_version(QuantumMachinesManager._get_local_proto_version())
            qop_proto_version = parse_proto_version(self._server_details.proto_version)

            if local_proto_version < qop_proto_version:
                logger.warning(
                    f"You are using an outdated version of `qm-qua` which was not tested against the current QOP "
                    f"version. Please consider updating to the latest qm-qua version, or at least to version"
                    f" '{qop_proto_version.base_version}'."
                )

    def version_dict(self) -> Version:
        """
        Returns:
            A dictionary with the qm-qua and QOP versions
        """
        warnings.warn(
            deprecation_message(
                "qmm.version_dict()",
                "1.3.0",
                "2.0.0",
                "Please use qmm.version() instead.",
            ),
            category=DeprecationWarning,
            stacklevel=2,
        )

        output_dict: Version = {}
        server_version = self._server_details.server_version
        output_dict["qm-qua"] = qm_qua_version
        qop_version = _get_qop_version(server_version)
        if qop_version is not None:
            output_dict["QOP"] = qop_version
        else:
            output_dict["OPX"] = server_version
        if is_debug():
            logger.debug(f"OPX version: {server_version}")

        return output_dict

    def version(self) -> DevicesVersion:
        """
        Returns:
            An object with the qm-qua and QOP versions
        """
        octaves = {}
        if self._octave_config is not None:
            for octave_name in self._octave_config.get_devices():
                octaves[octave_name] = self._octave_manager.get_client(octave_name).get_version()

        if self._api is None:
            gateway = self._server_details.server_version
            controllers = set(self._get_controllers_as_dict())
        else:
            response = self._api.get_version()
            gateway = response.gateway
            controllers = set(response.controllers)
        return DevicesVersion(
            gateway=gateway,
            controllers={k: None for k in controllers},  # type: ignore[misc]
            qm_qua=qm_qua_version,
            octaves=octaves,
        )

    def reset_data_processing(self) -> None:
        """
        Stops current data processing for all running jobs.

        Warning:
            Resetting the data processing is a disruptive operation.
            It should be used only as a recovery measure when the system becomes unresponsive due to heavy data processing, and is not intended for routine use.
        """
        if self._api:
            self._api.reset_data_processing()
        else:
            self._frontend.reset_data_processing()

    def open_qm(
        self,
        config: Union[FullQuaConfig, ControllerQuaConfig],
        close_other_machines: Optional[bool] = None,
        validate_with_protobuf: bool = False,
        add_calibration_elements_to_config: bool = True,
        use_calibration_data: bool = True,
        keep_dc_offsets_when_closing: bool = False,
        **kwargs: Any,
    ) -> Union[QuantumMachine, QmApi]:
        """Opens a new quantum machine. A quantum machine can use multiple OPXes, and a
        single OPX can also be used by multiple quantum machines as long as they do not
        share the same physical resources (input/output ports) as defined in the config.

        -- Available from QOP 3.5 --
        The configuration is split into two: controller config and logical config.
        When opening a QuantumMachine, the physical configuration defines the physical resources (e.g., ports), idle values (e.g., DC offsets), and port configurations.
        A full configuration (containing both logical and controller configs) can also be supplied, which will be used as the program's default.
        See the documentation website for more information.

        Args:
            config: The config that will be used by the Quantum Machine
            close_other_machines: When set to true, any open
                quantum machines will be closed. This simplifies the
                workflow but does not enable opening more than one
                quantum machine. The default `None` behavior is currently
                the same as `True`.
            validate_with_protobuf (bool): Validates config with
                protobuf instead of marshmallow. It is usually faster
                when working with large configs. Defaults to False.
            add_calibration_elements_to_config: Automatically adds config entries to allow Octave calibration.
            use_calibration_data: Automatically load updated calibration data from the calibration database into the config.
            keep_dc_offsets_when_closing: Available in QOP 2.4.2 - When closing the QM, do not change the DC offsets.

        Returns:
            A quantum machine obj that can be used to execute programs
        """
        if kwargs:
            logger.warning(f"unused kwargs: {list(kwargs)}, please remove them.")

        loaded_config = self._load_config(config, disable_marshmallow_validation=validate_with_protobuf)

        if use_calibration_data and self._octave_config is not None and self._octave_config.devices:
            if self._octave_config.calibration_db is not None:
                loaded_config = load_config_from_calibration_db(
                    loaded_config, self._octave_config.calibration_db, self._octave_config, self._caps
                )
            else:
                logger.warning("No calibration_db set in octave config, skipping loading calibration data")

        if add_calibration_elements_to_config and self._octave_config is not None and self._octave_config.devices:
            loaded_config = prep_config_for_calibration(loaded_config, self._octave_config, self._caps)

        octaves_config = get_controller_pb_config(loaded_config).octaves
        self._octave_manager.set_octaves_from_qua_config(octaves_config)

        if self._api:
            if keep_dc_offsets_when_closing:
                raise NotImplementedError("keep_dc_offsets_when_closing is not supported in the current QOP version")
            return self._api.open_qm(loaded_config, close_other_machines)

        close_other_machines = True if close_other_machines is None else close_other_machines

        if not self._caps.supports(QopCaps.keeping_dc_offsets):
            if keep_dc_offsets_when_closing:
                raise QmmException(
                    "The server does not support keeping DC offsets when closing. "
                    "Please upgrade the server to a version that supports this feature."
                )
        machine_id = self._frontend.open_qm(
            loaded_config, close_other_machines, keep_dc_offsets_when_closing=keep_dc_offsets_when_closing
        )

        return QuantumMachine(
            machine_id=machine_id,
            pb_config=loaded_config,
            frontend_api=self._frontend,
            capabilities=self._caps,
            octave_config=self._octave_config,
            octave_manager=self._octave_manager,
        )

    def _load_config(
        self, qua_config: Union[FullQuaConfig, ControllerQuaConfig], *, disable_marshmallow_validation: bool = False
    ) -> inc_qua_config_pb2.QuaConfig:
        try:
            if disable_marshmallow_validation:
                loaded_config = DictToQuaConfigConverter(self.capabilities).convert(qua_config)
            else:
                loaded_config = load_config(qua_config, capabilities=self.capabilities)
            validate_config_capabilities(loaded_config, self._caps)
            return loaded_config
        except KeyError as key_error:
            raise ConfigValidationException(f"Missing key {key_error} in config") from key_error
        except marshmallow.exceptions.ValidationError as validation_error:
            raise ConfigSchemaError(validation_error) from validation_error

    def validate_qua_config(self, qua_config: Union[FullQuaConfig, ControllerQuaConfig]) -> None:
        """
        Validates a qua config based on the connected server's capabilities.
        Raises an exception if the config is invalid.

        Args:
            qua_config: A python dict containing the qua config to validate
        """
        self._load_config(qua_config)

    def open_qm_from_file(self, filename: str, close_other_machines: bool = True) -> Union[QuantumMachine, QmApi]:
        """Opens a new quantum machine with config taken from a file on the local file system

        Args:
            filename: The path to the file that contains the config
            close_other_machines: Flag whether to close all other
                running machines

        Returns:
            A quantum machine obj that can be used to execute programs
        """
        warnings.warn(
            deprecation_message(
                "qmm.open_qm_from_file",
                "1.2.0",
                "2.0.0",
                "This method is going to be removed.",
            ),
            DeprecationWarning,
            stacklevel=1,
        )
        with open(filename) as json_file:
            json1_str = json_file.read()

            def remove_nulls(d: Dict[Any, Any]) -> Dict[Any, Any]:
                return {k: v for k, v in d.items() if v is not None}

            config = json.loads(json1_str, object_hook=remove_nulls)
        return self.open_qm(config, close_other_machines)

    def _create_simulation_request(
        self,
        config: FullQuaConfig,
        program: Program,
        simulate: SimulationConfig,
        compiler_options: Optional[CompilerOptionArguments] = None,
        *,
        strict: Optional[bool] = None,
        flags: Optional[List[str]] = None,
    ) -> frontend_pb2.SimulationRequest:
        """Creates a simulate request
        Args:
        """
        standardized_options = standardize_compiler_params(compiler_options, strict, flags)
        pb_config = load_config(config)

        return create_simulation_request(pb_config, program, simulate, standardized_options, self._caps)

    def simulate(
        self,
        config: FullQuaConfig,
        program: Program,
        simulate: SimulationConfig,
        compiler_options: Optional[CompilerOptionArguments] = None,
        *,
        strict: Optional[bool] = None,
        flags: Optional[List[str]] = None,
    ) -> Union[SimulatedJob, SimulatedJobApi]:
        """Simulate the outputs of a deterministic QUA program.

        The following example shows a simple execution of the simulator, where the
        associated config object is omitted for brevity.

        Example:
            ```python
            from qm.qua import *
            from qm import SimulationConfig, QuantumMachinesManager

            qmm = QuantumMachinesManager()

            with program() as prog:
                play('pulse1', 'qe1')

            job = qmm.simulate(config, prog, SimulationConfig(duration=100))
            ```

        Args:
            config: The full QUA configuration used to simulate the program, containing both the controller and logical configurations.
            program: A QUA ``program()`` object to execute
            simulate: A ``SimulationConfig`` configuration object
            compiler_options: additional parameters to pass to execute
            strict: a deprecated option for the compiler
            flags: deprecated way to provide flags to the compiler

        Returns:
            a ``QmJob`` object (see Job API).
        """
        self._caps.validate(program.used_capabilities)

        if self._api:
            request = self._create_simulation_request(
                config, program, simulate, compiler_options, strict=strict, flags=flags
            )

            simulated_job = self._api.simulate(
                request.config, request.highLevelProgram, request.simulate, request.controllerConnections
            )
            return simulated_job
        else:
            standardized_options = standardize_compiler_params(compiler_options, strict, flags)
            pb_config = load_config(config)
            job_id, simulated_response_part = self._simulation_api.simulate(
                pb_config, program, simulate, standardized_options, self._caps
            )
            return SimulatedJob(
                job_id=job_id,
                frontend_api=self._frontend,
                capabilities=self._server_details.capabilities,
                simulated_response=simulated_response_part,
            )

    def list_open_qms(self) -> List[str]:
        """Return a list of open quantum machines. (Returns only the ids, use ``get_qm(...)`` to get the machine object)

        Returns:
            The ids list
        """
        if self._api:
            return self._api.list_open_qms()
        else:
            return self._frontend.list_open_quantum_machines()

    def list_open_quantum_machines(self) -> List[str]:
        warnings.warn(
            deprecation_message(
                "qmm.list_open_quantum_machines",
                "1.2.0",
                "2.0.0",
                "This method was renamed to `qmm.list_open_qms()`",
            )
        )
        return self.list_open_qms()

    def get_qm(self, machine_id: str) -> Union[QuantumMachine, QmApi]:
        """Gets an open quantum machine object with the given machine id

        Args:
            machine_id: The id of the open quantum machine to get

        Returns:
            A quantum machine obj that can be used to execute programs
        """
        if self._api:
            return self._api.get_qm(machine_id)

        qm_data = self._frontend.get_quantum_machine(machine_id)
        machine_id = qm_data.machine_id
        pb_config = qm_data.config

        return QuantumMachine(
            machine_id=machine_id,
            pb_config=pb_config,
            frontend_api=self._frontend,
            capabilities=self._caps,
            octave_manager=self.octave_manager,
        )

    def get_job_result_handles(self, job_id: str) -> StreamsManager:
        """
        -- Available in QOP 2.x --

        Returns the result handles for a job.

        Args:
            job_id: The job id

        Returns:
            The handles that this job generated
        """
        if self._api is not None:
            raise NotImplementedError(
                "This method is not available in the current QOP version, "
                "please use `qmm.get_job(job_id).result_handles`"
            )
        return StreamsManager(
            JobResultServiceApi(self._server_details.connection_details, job_id), self._caps, wait_until_func=None
        )

    def close_all_qms(self) -> None:
        """Closes ALL open quantum machines"""
        if self._api:
            self._api.close_all_qms()
        else:
            self._frontend.close_all_quantum_machines()

    def close_all_quantum_machines(self) -> None:
        warnings.warn(
            deprecation_message(
                "qmm.close_all_quantum_machines",
                "1.2.0",
                "2.0.0",
                "This function is going to change its name to `qmm.close_all_qms()`",
            ),
            category=DeprecationWarning,
            stacklevel=2,
        )
        self.close_all_qms()

    def get_controllers(self) -> List[ControllerBase]:
        """Returns a list of all the controllers that are available"""
        if self._api:
            warnings.warn(
                deprecation_message(
                    "get_controllers",
                    "1.2.0",
                    "2.0.0",
                    "This will have a different return type",
                ),
                category=DeprecationWarning,
                stacklevel=2,
            )
            new_response = self._api.get_controllers()
            return list(new_response.values())
        else:
            old_response = self._frontend.get_controllers()
            return [
                Controller(message.name, message.temperature if message.HasField("temperature") else None)
                for message in old_response
            ]

    def _get_controllers_as_dict(self) -> Mapping[str, ControllerBase]:
        if self._api:
            return self._api.get_controllers()
        else:
            old_response = self._frontend.get_controllers()
            return {
                message.name: Controller(message.name, message.temperature if message.HasField("temperature") else None)
                for message in old_response
            }

    def get_devices(self) -> Devices:
        controllers = self._get_controllers_as_dict()
        octaves: Dict[str, OctaveDetails] = {}
        if self._octave_config is not None:
            for octave_name in self._octave_config.devices:
                octave_client = self._octave_manager.get_client(octave_name)
                octaves[octave_name] = octave_client.get_details()
        return Devices(controllers=controllers, octaves=octaves)

    def clear_all_job_results(self) -> None:
        """Deletes all data from all previous jobs"""
        if self._api:
            warnings.warn(
                deprecation_message(
                    "clear_all_job_results",
                    "1.2.0",
                    "2.0.0",
                    "This method is going to be removed.",
                ),
                category=DeprecationWarning,
                stacklevel=1,
            )
            self._api.clear_all_job_results()
        else:
            self._frontend.clear_all_job_results()

    def get_jobs(
        self,
        qm_ids: Iterable[str] = tuple(),
        job_ids: Iterable[str] = tuple(),
        user_ids: Iterable[str] = tuple(),
        description: str = "",
        status: Iterable[JobStatus] = tuple(),
    ) -> List[JobData]:
        """
        -- Available in QOP 3.x --

        Get jobs based on filtering criteria. All fields are optional.

        Args:
            qm_ids: A list of qm ids
            job_ids: A list of jobs ids
            user_ids: A list of user ids
            description: Jobs' description
            status: A list of job statuses

        Returns:
            A list of jobs
        """
        if self._api is None:
            raise NotImplementedError("This method is not available in the current QOP version")
        return self._api.get_jobs(
            qm_ids=qm_ids, job_ids=job_ids, user_ids=user_ids, description=description, status=status
        )

    def get_job(self, job_id: str) -> JobApi:
        """
        -- Available in QOP 3.x --

        Get a job based on the job_id.

        Args:
            job_id: A list of jobs ids

        Returns:
            The job
        """
        if self._api is None:
            raise NotImplementedError("This method is not available in the current QOP version")
        return self._api.get_job(job_id)

    @property
    def _debug_data(self) -> Optional[DebugData]:
        return self._server_details.connection_details.debug_data

    @staticmethod
    def set_capabilities_offline(
        capabilities: Optional[Union[Collection[Capability], ServerCapabilities]] = None,
    ) -> None:
        """
        Some modules of the sdk cannot be run without the capabilities of the QOP server, which is automatically set
        when connecting to the server via QuantumMachinesManager (in "_initialize_connection"). This function provides
        an alternative to connecting to a QOP server via QuantumMachinesManager, by setting the capabilities
        manually and globally.

        It is possible to extract the capabilities of an existing QuantumMachinesManager object by checking the `capabilities` attribute,
        e.g. `qmm = QuantumMachinesManager(); capabilities = qmm.capabilities`.

        The QopCaps class (from `qm import QopCaps`) can be used to get and view all capabilities.

        Warning: Setting the capabilities when there is an open QuantumMachinesManager will override them and can cause unexpected behavior.

        Args:
            capabilities: A set of capabilities (or a ServerCapabilities object) to create the container with. If None,
                all capabilities are set. Warning: Using all capabilities might not produce the expected behavior.

        """
        if capabilities is None:
            capabilities = QopCaps.get_all()
        elif isinstance(capabilities, ServerCapabilities):
            capabilities = capabilities.supported_capabilities

        offline_capabilities.set(capabilities)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: io/qualang/api/v1/info_service.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'io/qualang/api/v1/info_service.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n$io/qualang/api/v1/info_service.proto\x12\x11io.qualang.api.v1\"\x10\n\x0eGetInfoRequest\"i\n\x0fGetInfoResponse\x12@\n\x0eimplementation\x18\x01 \x01(\x0b\x32(.io.qualang.api.v1.ImplementationDetails\x12\x14\n\x0c\x63\x61pabilities\x18\x02 \x03(\t\"q\n\x15ImplementationDetails\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07version\x18\x02 \x01(\t\x12\x0b\n\x03url\x18\x03 \x01(\t\x12\x1a\n\rproto_version\x18\x04 \x01(\tH\x00\x88\x01\x01\x42\x10\n\x0e_proto_version2a\n\x0bInfoService\x12R\n\x07GetInfo\x12!.io.qualang.api.v1.GetInfoRequest\x1a\".io.qualang.api.v1.GetInfoResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'io.qualang.api.v1.info_service_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_GETINFOREQUEST']._serialized_start=59
  _globals['_GETINFOREQUEST']._serialized_end=75
  _globals['_GETINFORESPONSE']._serialized_start=77
  _globals['_GETINFORESPONSE']._serialized_end=182
  _globals['_IMPLEMENTATIONDETAILS']._serialized_start=184
  _globals['_IMPLEMENTATIONDETAILS']._serialized_end=297
  _globals['_INFOSERVICE']._serialized_start=299
  _globals['_INFOSERVICE']._serialized_end=396
# @@protoc_insertion_point(module_scope)



import logging
from typing import List, Type, Mapping, Iterator, Optional

from google.protobuf.wrappers_pb2 import Int64Value

from qm.api.base_api import BaseApi
from qm.grpc.qm.pb import job_results_pb2
from qm.api.models.server_details import ConnectionDetails
from qm.utils.protobuf_utils import proto_repeated_to_list
from qm.grpc.qm.pb.job_results_pb2_grpc import JobResultsServiceStub
from qm.api.models.jobs import JobNamedResult, JobStreamingState, JobResultItemSchema, JobNamedResultHeader

logger = logging.getLogger(__name__)


class JobResultServiceApi(BaseApi[JobResultsServiceStub]):
    def __init__(self, connection_details: ConnectionDetails, job_id: str):
        super().__init__(connection_details)
        self._id = job_id

    @property
    def id(self) -> str:
        return self._id

    @property
    def _stub_class(self) -> Type[JobResultsServiceStub]:
        return JobResultsServiceStub

    def get_job_errors(self) -> List[job_results_pb2.GetJobErrorsResponse.Error]:
        request = job_results_pb2.GetJobErrorsRequest(jobId=self._id)
        response: job_results_pb2.GetJobErrorsResponse = self._run(
            self._stub.GetJobErrors, request, timeout=self._timeout
        )
        return proto_repeated_to_list(response.errors)

    def get_job_named_result(
        self,
        output_name: str,
        long_offset: int,
        limit: int,
        timeout: Optional[float],
    ) -> Iterator[JobNamedResult]:
        timeout = timeout if timeout is not None else self._timeout

        request = job_results_pb2.GetJobNamedResultRequest(
            jobId=self._id, outputName=output_name, longOffset=Int64Value(value=long_offset), limit=limit
        )
        response = self._run_iterator(self._stub.GetJobNamedResult, request, timeout=timeout)

        for result in response:
            yield JobNamedResult(data=result.data, count_of_items=result.countOfItems, output_name=output_name)

    def get_job_state(self) -> JobStreamingState:
        request = job_results_pb2.GetJobStateRequest(jobId=self._id)
        response = self._run(self._stub.GetJobState, request, timeout=self._timeout)
        return JobStreamingState(
            job_id=self._id,
            done=response.done,
            closed=response.closed,
            has_dataloss=response.hasDataloss,
        )

    def get_named_header(self, output_name: str, flat_struct: bool) -> JobNamedResultHeader:
        request = job_results_pb2.GetJobNamedResultHeaderRequest(
            jobId=self._id, outputName=output_name, flatFormat=flat_struct
        )
        response = self._run(self._stub.GetJobNamedResultHeader, request, timeout=self._timeout)
        return JobNamedResultHeader(
            count_so_far=response.countSoFar,
            bare_dtype=response.simpleDType,
            shape=tuple(response.shape),
            has_dataloss=response.hasDataloss,
            has_execution_errors=response.hasExecutionErrors.value,
        )

    def get_state_from_header(self, output_name: str, flat_struct: bool) -> JobStreamingState:
        """This function is for backward compatibility, when we didn't have the get_job_state"""
        request = job_results_pb2.GetJobNamedResultHeaderRequest(
            jobId=self._id, outputName=output_name, flatFormat=flat_struct
        )
        response: job_results_pb2.GetJobNamedResultHeaderResponse = self._run(
            self._stub.GetJobNamedResultHeader, request, timeout=self._timeout
        )
        return JobStreamingState(
            job_id=self._id,
            done=response.done,
            closed=response.closed,
            has_dataloss=response.hasDataloss,
        )

    def get_job_result_schema(self) -> Mapping[str, JobResultItemSchema]:
        request = job_results_pb2.GetJobResultSchemaRequest(jobId=self._id)
        response = self._run(self._stub.GetJobResultSchema, request, timeout=self._timeout)
        return {
            item.name: JobResultItemSchema(
                name=item.name,
                bare_dtype=item.simpleDType,
                shape=tuple(item.shape),
                is_single=item.isSingle,
                expected_count=item.expectedCount,
            )
            for item in response.items
        }



from typing import Any, List, Union, Mapping, TypeVar, Iterator, Sequence

import numpy as np

from qm.type_hinting.general import Number
from qm.qua._expressions import QuaVariable

V = TypeVar("V")
T = TypeVar("T", float, int)

QuaIterableArrayType = Sequence[T]
AnyQuaIterableArrayType = Union[QuaIterableArrayType[float], QuaIterableArrayType[int]]
MetaDataType = Mapping[str, Any]
QuaIterableArrayInputInt = Union[np.typing.NDArray[np.integer[Any]], QuaIterableArrayType[int]]
QuaIterableArrayInputFloat = Union[np.typing.NDArray[np.floating[Any]], QuaIterableArrayType[float]]
QuaIterableArrayInput = Union[QuaIterableArrayInputInt, QuaIterableArrayInputFloat]
NativeArrayType = Sequence[Any]

QuaIterablesDType = type[Number]
MultiIteratorsValuesType = List[Union[NativeArrayType, AnyQuaIterableArrayType]]

NativeIteratorItemType = Any  # Supporting any iterable for native type
QuaVariableTypes = Union[QuaVariable[float], QuaVariable[int]]

IteratorContentTypes = Union[
    NativeArrayType, AnyQuaIterableArrayType, NativeIteratorItemType, QuaVariableTypes, "QuaNamedTuple"
]

# iterator types
QuaIteratorType = Iterator[QuaVariable[T]]
AnyQuaIteratorType = Union[QuaIteratorType[float], QuaIteratorType[int]]
MultiIteratorType = Iterator["QuaNamedTuple"]
NativeIteratorType = Iterator[V]
IteratorType = Union[AnyQuaIteratorType, MultiIteratorType, NativeIteratorType[Any]]


class QuaNamedTuple:
    def __init__(self, names: Sequence[str], values: Sequence[Any]) -> None:
        self._fields = tuple(names)
        for key, value in zip(names, values):
            setattr(self, key, value)
        self._frozen = True

    # The -> Any return type hint tells IDEs that dynamic attribute access is valid,
    # suppressing "unresolved attribute" warnings for fields set via setattr in __init__.
    def __getattr__(self, key: str) -> Any:
        raise AttributeError(f"'{self.__class__.__name__}' has no attribute '{key}'")

    def __setattr__(self, key: str, value: Any) -> None:
        if getattr(self, "_frozen", False):
            raise AttributeError(f"Cannot modify attribute '{key}' of {self.__class__.__name__}")
        super().__setattr__(key, value)

    def __iter__(self) -> MultiIteratorType:
        for field in self._fields:
            yield getattr(self, field)

    def __repr__(self) -> str:
        values = ", ".join(f"{f}={getattr(self, f)!r}" for f in self._fields)
        return f"{self.__class__.__name__}({values})"

    @property
    def name(self) -> str:
        return "_".join(self._fields)



def format_float(f: float) -> str:
    return "{:.3f}".format(f)


def pretty_string_freq(f: float) -> str:
    if f < 1000:
        div, units = 1.0, "Hz"
    elif 1000 <= f < 1_000_000:
        div, units = 1000.0, "kHz"
    else:
        div, units = 1e6, "MHz"
    return f"{format_float(f / div).rstrip('0').rstrip('.')}{units}"
